/*
 *  asus_acpi.c - Asus Laptop ACPI Extras
 *
 *
 *  Copyright (C) 2002-2005 Julien Lerouge, 2003-2006 Karol Kozimor
 *  Copyright (C) 2006-2007 Corentin Chary
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *  The development page for this driver is located at
 *  http://sourceforge.net/projects/acpi4asus/
 *
 *  Credits:
 *  Pontus Fuchs   - Helper functions, cleanup
 *  Johann Wiesner - Small compile fixes
 *  John Belmonte  - ACPI code for Toshiba laptop was a good starting point.
 *  Eric Burghard  - LED display support for W1N
 *  Josh Green     - Light Sens support
 *  Brice Arnould  - Some help for parse_method()
 *
 */

#include <linux/version.h>

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,18)
# include <linux/config.h>
#else
# include <linux/autoconf.h>
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/err.h>
#include <linux/proc_fs.h>
#include <linux/backlight.h>
#include <acpi/acpi_drivers.h>
#include <acpi/acpi_bus.h>
#include <asm/uaccess.h>

#if defined(CONFIG_BACKLIGHT_CLASS_DEVICE) || defined(CONFIG_BACKLIGHT_CLASS_DEVICE_MODULE)
# define CONFIG_ACPI_ASUS_NEW_BACKLIGHT
#endif


#define ASUS_ACPI_VERSION "0.32"

#define PROC_ASUS       "asus"	//the directory
#define PROC_MLED       "mled"
#define PROC_WLED       "wled"
#define PROC_TLED       "tled"
#define PROC_RLED       "rled"
#define PROC_BT         "bluetooth"
#define PROC_LEDD       "ledd"
#define PROC_INFO       "info"
#define PROC_LCD        "lcd"
#define PROC_BRN        "brn"
#define PROC_DISP       "disp"
#define PROC_LSSW       "lssw"
#define PROC_LSLVL      "lslvl"

#define ACPI_HOTK_NAME          "Asus Laptop ACPI Extras Driver"
#define ACPI_HOTK_CLASS         "hotkey"
#define ACPI_HOTK_DEVICE_NAME   "Hotkey"
#define ACPI_HOTK_HID           "ATK0100"

/*
 * Some events we use, same for all Asus
 */
#define BR_UP       0x10
#define BR_DOWN     0x20

/*
 * Flags for hotk status
 */
#define MLED_ON     0x01	//mail LED
#define WLED_ON     0x02	//wireless LED
#define TLED_ON     0x04	//touchpad LED
#define BT_ON       0x08	//internal Bluetooth
#define RLED_ON     0x10        //Record LED

/*
 * Methods configuration
 */
#define MT_MAX_ARGS 3

MODULE_AUTHOR("Julien Lerouge, Karol Kozimor, Corentin Chary");
MODULE_DESCRIPTION(ACPI_HOTK_NAME);
MODULE_LICENSE("GPL");

static uint unsupported;
module_param(unsupported, uint, 0);
MODULE_PARM_DESC(unsupported, "Allow the use of default values for an unsupported model.\n");

/* For each model, all features implemented,
 * those marked with R are relative to HOTK, A for absolute.
 */
struct model_data {
	char *name;		//name of the laptop________________A
	char *mled_set;		//method to handle mled_____________R
	char *mled_status;	//node to handle mled reading_______A
	char *wled_set;		//method to handle wled_____________R
	char *wled_status;	//node to handle wled reading_______A
	char *rled_set;         //method to handle rled_____________R
        char *rled_status;      //node to handle rled reading_______A
	char *tled_set;		//method to handle tled_____________R
	char *tled_status;	//node to handle tled reading_______A
	char *ledd_set;		//method to handle LED display______R
	char *bt_switch;	//method to switch Bluetooth on/off_R
	char *bt_status;	//method to read Bluetooth state____A
	char *lcd_switch;	//method to turn LCD on/off_________A
	char *lcd_status;	//node to read LCD panel state______A
	char *brightness_up;	//method to set brightness up_______A
	char *brightness_down;	//guess what ?______________________A
	char *brightness_set;	//method to set absolute brightness_R
	char *brightness_get;	//method to get absolute brightness_R
	char *brightness_status; //node to get brightness____________A
	char *light_sens_switch; //light sensor toggle method________R
	char *light_sens_level; //light sensor level set method_____R
	char *display_set;	//method to set video output________R
	char *display_get;	//method to get video output________R
};

/* A method have a name, and from 0 to MT_MAX_ARGS params ... */
struct method {
	const char * name;
	struct acpi_object_list input;
	union acpi_object params[MT_MAX_ARGS];
};

/*
 * This is the main structure, we can use it to store anything interesting
 * about the hotk device
 */
struct asus_hotk {
	struct acpi_device *device;	//the device we are in
	acpi_handle handle;	//the handle of the hotk device
	char status;		//status of the hotk, for LEDs, ...
	u32 ledd_status;	//status of the LED display
	struct model_data *methods;	//methods available on the laptop
	u8 brightness;		//brightness level
	u8 light_level;    //light sensor level
	u8 light_switch;    //light sensor switch value
	enum {
		A1x = 0,	//A1340D, A1300F
		A2D,            //A2D
		A2x,		//A2500H
		A3E,            //A3E
		A3F,            //A3F, A6F
		A3N,            //A3N, A3L
		A4G,		//A4700G
		A4K,            //A4K, A4D
		A4L,            //A4L
		A6B,            //A6B, A6K, A6U
		A6J,            //A6J
		A6R,            //A6R
		A6T,            //A6T, A6M
		A6V,            //A6V, A6Q
		A7D,            //A7D
		A7J,            //A7J
		A7T,            //A7T
		A8J,            //A8J
		D1x,		//D1
		F3J,            //F3JM
		F3T,            //F3T
		L2D,		//L2000D
		L3C,		//L3800C
		L3D,		//L3400D
		L3H,		//L3H, L2000E, L5D
		L4R,		//L4500R
		L5x,		//L5800C
		L8L,		//L8400L
		M1A,		//M1300A
		M2E,		//M2400E, L4400L
		M6N,		//M6800N, W3400N
		M6R,		//M6700R, A3000G
		M6V,            //M6A, M6V-B022P, M6V-8012P
		M7A,            //Z71A
		M7V,            //Z71V
		M9V,            //M9V
		P30,		//Samsung P30
		S1x,		//S1300A, but also L1400B and M2400A (L84F)
		S2x,		//S200 (J1 reported), Victor MP-XP7210
		S5A,            //S5200A, M5200A, z33A
		U5F,            //U5F
		VX1,            //VX-1
		V6J,            //V6J
		V6V,            //V6V
		W1J,            //W1Jc
		W1N,		//W1000N
		W2V,            //W2V
		W3V,            //W3030V
		W3Z,            //W3Z
		W5A,		//W5A
		W5F,            //W5F
		W6A,            //W6A
		xxN,		//M2400N, M3700N, M5200N, M6800N, S1300N, S5200N
		Z96,            //Z96F
		//(Centrino)
		DEF,            //Default values
		END_MODEL
	} model;		//Models currently supported
	u16 event_count[128];	//count for each event TODO make this better
};

/* Here we go */
#define A1x_PREFIX "\\_SB.PCI0.ISA.EC0."
#define L3C_PREFIX "\\_SB.PCI0.PX40.ECD0."
#define M1A_PREFIX "\\_SB.PCI0.PX40.EC0."
#define P30_PREFIX "\\_SB.PCI0.LPCB.EC0."
#define S1x_PREFIX "\\_SB.PCI0.PX40."
#define S2x_PREFIX A1x_PREFIX
#define xxN_PREFIX "\\_SB.PCI0.SBRG.EC0."
#define A6x_PREFIX xxN_PREFIX

static struct model_data model_conf[END_MODEL] = {
	/*
	 * TODO I have seen a SWBX and AIBX method on some models, like L1400B,
	 * it seems to be a kind of switch, but what for ?
	 */

	{
	 .name = "A1x",
	 .mled_set = "MLED",
	 .mled_status = "\\MAIL",
	 .lcd_switch = A1x_PREFIX "_Q10",
	 .lcd_status = "\\BKLI",
	 .brightness_up = A1x_PREFIX "_Q0E",
	 .brightness_down = A1x_PREFIX "_Q0F"},

	{
	 .name = "A2D",
	 .mled_set = "MLED",
	 .wled_set = "WLED",
	 .lcd_switch = xxN_PREFIX "_Q10",
	 .lcd_status = "\\_SB.PCI0.SBRG.GO18",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\ACTD"},

	{
	 .name = "A2x",
	 .mled_set = "MLED",
	 .wled_set = "WLED",
	 .wled_status = "\\SG66",
	 .lcd_switch = "\\Q10",
	 .lcd_status = "\\BAOF",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
 	 .display_get = "\\INFB"},

	{
	 .name = "A3E",
	 .mled_set = "MLED",
	 .wled_set = "WLED",
	 .lcd_switch = xxN_PREFIX "_Q10",
	 .lcd_status = xxN_PREFIX "RPIN(0x67)",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\_SB.PCI0.P0P2.VGA.GETD"},

	{
	 .name = "A3F",
	 .mled_set = "MLED",
	 .wled_set = "WLED",
	 .wled_status = "\\WRST",
	 .bt_switch = "BLED",
	 .bt_status = "\\BRST",
	 .lcd_switch = xxN_PREFIX "_Q10",
	 .lcd_status = xxN_PREFIX "RPIN(0x11)",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\SSTE"},

	{
	 .name = "A3N",
	 .mled_set = "MLED",
	 .wled_set = "WLED",
	 .wled_status = "\\WLDC",
	 .lcd_switch = xxN_PREFIX "_Q10",
	 .lcd_status = "\\BKLT",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\SSTE"},

	{
	 .name = "A4G",
	 .mled_set = "MLED",
/* WLED present, but not controlled by ACPI */
	 .lcd_switch = xxN_PREFIX "_Q10",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\ADVG"},

	{
	 .name = "A4K",
	 .mled_set = "MLED",
/* WLED and TLED present, but not controlled by ACPI */
	 .lcd_switch = xxN_PREFIX "_Q10",
	 .lcd_status = "\\BLFG",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\_SB.PCI0.P0P2.VGA.GETD"},

	{
	 .name = "A4L",
	 .mled_set = "MLED",
/* WLED present, but not controlled by ACPI */
	 .lcd_switch = xxN_PREFIX "_Q10",
	 .lcd_status = "\\RIOP(0x45,0x01)",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\_SB.PCI0.P0P2.VGA.GETD"},

	{
	 .name = "A6B",
	 .mled_set = "MLED",
	 .wled_set = "WLED",
	 .lcd_switch = A6x_PREFIX "_Q10",
	 .lcd_status = "\\RGPL(0x07,0x01)",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\_SB.PCI0.P0P1.VGA.GETD"},

	{
	 .name = "A6J",
	 .mled_set = "MLED",
	 .wled_set = "WLED",
	 .wled_status = "\\WRST",
	 .bt_switch = "BLED",
	 .bt_status = "\\BRST",
	 .lcd_switch = A6x_PREFIX "_Q10",
	 .lcd_status = "\\RGP0(0x06)",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\_SB.PCI0.P0P2.VGA.GETD"},

	{
	 .name = "A6R",
	 .mled_set = "MLED",
	 .wled_set = "WLED",
	 .wled_status = "\\WRST",
	 .bt_switch = "BLED",
	 .bt_status = "\\BRST",
	 .lcd_switch = A6x_PREFIX "_Q10",
	 .lcd_status = A6x_PREFIX "RPIN(0x11)",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\_SB.PCI0.P0P1.VGA.GETD"},

	{
	 .name = "A6T",
	 .mled_set = "MLED",
	 .wled_set = "WLED",
	 .wled_status = "\\WWED",
	 .bt_switch = "BLED",
	 .bt_status = "\\BLTH",
	 .lcd_switch = A6x_PREFIX "_Q10",
	 .lcd_status = "\\BKLT",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\_SB.PCI0.P0PA.VGA.GETD"},

	{
	 .name = "A6V",
	 .mled_set = "MLED",
	 .wled_set = "WLED",
	 .bt_switch = "BLED",
	 .lcd_switch = A6x_PREFIX "_Q10",
	 .lcd_status = "\\RGPL(0x15,0x01)",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\_SB.PCI0.P0P3.VGA.GETD"},

	{
	 .name = "A7D",
	 .mled_set = "MLED",
	 .wled_set = "WLED",
	 .wled_status = "\\WLFG",
	 .bt_switch = "BLED",
	 .bt_status = "\\BLFG",
	 .lcd_switch = A6x_PREFIX "_Q10",
	 .lcd_status = "\\_SB.PCI0.SMBS.GPO7",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\_SB.PCI0.P0P1.VGA.GETD"},

	{
	 .name = "A7J",
	 .mled_set = "PLED",
	 .wled_set = "WLED",
	 .bt_switch = "BLED",
	 .lcd_switch = A6x_PREFIX "_Q10",
	 .lcd_status = A6x_PREFIX "RPIN(0x11)",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\_SB.PCI0.P0P2.VGA.GETD"},

	{
	 .name = "A7T",
	 .mled_set = "PLED",
	 .wled_set = "WLED",
	 .wled_status = "\\WWED",
	 .bt_switch = "BLED",
	 .bt_status = "\\BLTH",
	 .lcd_switch = A6x_PREFIX "_Q10",
	 .lcd_status = "\\BKLT",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\_SB.PCI0.P0PA.VGA.GETD"},

	{
	 .name = "A8J",
	 .mled_set = "MLED",
	 .wled_set = "WLED",
	 .wled_status = "\\WRST",
	 .bt_switch = "BLED",
	 .bt_status = "\\BRST",
	 .lcd_switch = A6x_PREFIX "_Q10",
	 .lcd_status = "\\RGPL(0x06,0x01)",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\_SB.PCI0.P0P2.VGA.GETD"},

	{
	 .name = "D1x",
	 .mled_set = "MLED",
	 .lcd_switch = "\\Q0D",
	 .lcd_status = "\\GP11",
	 .brightness_up = "\\Q0C",
	 .brightness_down = "\\Q0B",
	 .brightness_status = "\\BLVL",
	 .display_set = "SDSP",
	 .display_get = "\\INFB"},

	{
	 .name = "F3J",
	 .mled_set = "MLED",
	 .wled_set = "WLED",
	 .wled_status = "\\WRST",
	 .bt_switch = "BLED",
	 .bt_status = "\\BRST",
	 .tled_set = "TLED",
	 .lcd_switch = A6x_PREFIX "_Q10",
	 .lcd_status = A6x_PREFIX "RPIN(0x11)",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\_SB.PCI0.P0P1.VGA.GETD"},

	{
	 .name = "F3T",
	 .mled_set = "MLED",
	 .wled_set = "WLED",
	 .bt_switch = "BLED",
	 .lcd_switch = A6x_PREFIX "_Q10",
	 .lcd_status = A6x_PREFIX "RPIN(0x11)",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\_SB.PCI0.P0PA.VGA.GETD"},

	{
	 .name = "L2D",
	 .mled_set = "MLED",
	 .mled_status = "\\SGP6",
	 .wled_set = "WLED",
	 .wled_status = "\\RCP3",
	 .lcd_switch = "\\Q10",
	 .lcd_status = "\\SGP0",
	 .brightness_up = "\\Q0E",
	 .brightness_down = "\\Q0F",
	 .display_set = "SDSP",
	 .display_get = "\\INFB"},

	{
	 .name = "L3C",
	 .mled_set = "MLED",
	 .wled_set = "WLED",
	 .lcd_switch = L3C_PREFIX "_Q10",
	 .lcd_status = "\\GL32",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\_SB.PCI0.PCI1.VGAC.NMAP"},

	{
	 .name = "L3D",
	 .mled_set = "MLED",
	 .mled_status = "\\MALD",
	 .wled_set = "WLED",
	 .lcd_switch = "\\Q10",
	 .lcd_status = "\\BKLG",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\INFB"},

	{
	 .name = "L3H",
	 .mled_set = "MLED",
	 .wled_set = "WLED",
	 .lcd_switch = "EHK(0x07)",
	 .lcd_status = "\\_SB.PCI0.PM.PBC(0x02,0x02)",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\INFB"},

	{
	 .name = "L4R",
	 .mled_set = "MLED",
	 .wled_set = "WLED",
	 .wled_status = "\\_SB.PCI0.SBRG.SG13",
	 .lcd_switch = xxN_PREFIX "_Q10",
	 .lcd_status = "\\_SB.PCI0.SBSM.SEO4",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\_SB.PCI0.P0P1.VGA.GETD"},

	{
	 .name = "L5x",
	 .mled_set = "MLED",
/* WLED present, but not controlled by ACPI */
	 .tled_set = "TLED",
	 .lcd_switch = "\\Q0D",
	 .lcd_status = "\\BAOF",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\INFB"},

	{
	 .name = "L8L"
/* No features, but at least support the hotkeys */
	 },

	{
	 .name = "M1A",
	 .mled_set = "MLED",
	 .lcd_switch = M1A_PREFIX "Q10",
	 .lcd_status = "\\PNOF",
	 .brightness_up = M1A_PREFIX "Q0E",
	 .brightness_down = M1A_PREFIX "Q0F",
	 .brightness_status = "\\BRIT",
	 .display_set = "SDSP",
	 .display_get = "\\INFB"},

	{
	 .name = "M2E",
	 .mled_set = "MLED",
	 .wled_set = "WLED",
	 .lcd_switch = "\\Q10",
	 .lcd_status = "\\GP06",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\INFB"},

	{
	 .name = "M6N",
	 .mled_set = "MLED",
	 .wled_set = "WLED",
	 .wled_status = "\\_SB.PCI0.SBRG.SG13",
	 .lcd_switch = xxN_PREFIX "_Q10",
	 .lcd_status = "\\_SB.BKLT",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\SSTE"},

	{
	 .name = "M6R",
	 .mled_set = "MLED",
	 .wled_set = "WLED",
	 .lcd_switch = xxN_PREFIX "_Q10",
	 .lcd_status = "\\_SB.PCI0.SBSM.SEO4",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\_SB.PCI0.P0P1.VGA.GETD"},

	{
	 .name = "M6V",
	 .mled_set = "MLED",
	 .wled_set = "WLED",
	 .wled_status = "\\WLSF",
	 .bt_switch = "BLED",
	 .bt_status = "\\BTST",
	 .lcd_switch = xxN_PREFIX "_Q10",
	 .lcd_status = "\\RGPL(0x15,0x01)",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\_SB.PCI0.P0P1.VGA.GETD"},

	{
	 .name = "M7A",
	 .mled_set = "MLED",
	 .lcd_switch = xxN_PREFIX "_Q10",
	 .lcd_status = xxN_PREFIX "RPIN(0x56)",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .light_sens_switch = "ALSC",
	 .light_sens_level = "ALSL",
	 .display_set = "SDSP",
	 .display_get = "\\ADVG"},

	{
	 .name = "M7V",
	 .mled_set = "MLED",
	 .lcd_switch = xxN_PREFIX "_Q10",
	 .lcd_status = xxN_PREFIX "RPIN(0x56)",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .light_sens_switch = "ALSC",
	 .light_sens_level = "ALSL",
	 .display_set = "SDSP",
	 .display_get = "\\_SB.PCI0.P0P2.VGA.GETD"},

        {
         .name = "M9V",
         .wled_set = "WLED",
	 .wled_status = "\\WLST",
         .bt_switch = "BLED",
	 .bt_status = "\\BTST",
         .lcd_switch = xxN_PREFIX "_Q10",
         .lcd_status = "\\RGPL(0x15,0x01)",
         .brightness_set = "SPLV",
         .brightness_get = "GPLV",
         .display_set = "SDSP",
         .display_get = "\\_SB.PCI0.P0P2.VGA.GETD"},

	{
	 .name = "P30",
	 .wled_set = "WLED",
	 .lcd_switch = P30_PREFIX "_Q0E",
	 .lcd_status = "\\BKLT",
	 .brightness_up = P30_PREFIX "_Q68",
	 .brightness_down = P30_PREFIX "_Q69",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\DNXT"},

	{
	 .name = "S1x",
	 .mled_set = "MLED",
	 .mled_status = "\\EMLE",
	 .wled_set = "WLED",
	 .lcd_switch = S1x_PREFIX "Q10",
	 .lcd_status = "\\PNOF",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV"},

	{
	 .name = "S2x",
	 .mled_set = "MLED",
	 .mled_status = "\\MAIL",
	 .lcd_switch = S2x_PREFIX "_Q10",
	 .lcd_status = "\\BKLI",
	 .brightness_up = S2x_PREFIX "_Q0B",
	 .brightness_down = S2x_PREFIX "_Q0A"},

        {
         .name = "S5A",
         .mled_set = "MLED",
         .wled_set = "WLED",
         .lcd_switch = xxN_PREFIX "_Q10",
         .lcd_status = "\\RGPL(0x15,0x01)",
         .brightness_set = "SPLV",
         .brightness_get = "GPLV",
         .display_set = "SDSP",
         .display_get = "\\_SB.PCI0.P0P2.VGA.GETD"},

        {
         .name = "U5F",
         .tled_set = "TLED",
         .wled_set = "WLED",
	 .wled_status = "\\WLNS",
	 .bt_switch = "BLED",
	 .bt_status = "\\WBTS",
         .lcd_switch = xxN_PREFIX "_Q10",
         .lcd_status = "\\RGPL(0x09,0x01)",
         .brightness_set = "SPLV",
         .brightness_get = "GPLV",
         .display_set = "SDSP",
         .display_get = "\\ADVG"},

        {
         .name = "VX1",
         .tled_set = "TLED",
         .wled_set = "WLED",
	 .bt_switch = "BLED",
         .lcd_switch = xxN_PREFIX "_Q10",
         .lcd_status = "\\RGPL(0x06,0x01)",
         .brightness_set = "SPLV",
         .brightness_get = "GPLV",
         .display_set = "SDSP",
         .display_get = "\\_SB.PCI0.P0P1.VGA.GETD"},

	{
	 .name = "V6J",
	 .wled_set = "WLED",
	 .bt_switch = "BLED",
	 .tled_set = "TLED",
	 .lcd_switch = xxN_PREFIX "_Q10",
	 .lcd_status = "\\RGPL(0x06,0x01)",
	 .display_get = "\\_SB.PCI0.P0P1.VGA.GETD",
	 .display_set = "SDSP",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV"},

	{
	 .name = "V6V",
	 .bt_switch = "BLED",
	 .bt_status = "\\_SB.PCI0.SBRG.BLTS",
	 .wled_set = "WLED",
	 .tled_set = "TLED",
	 .lcd_switch = xxN_PREFIX "_Q10",
	 .lcd_status = "\\BKLT",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_get = "\\_SB.PCI0.P0P1.VGA.GETD",
	 .display_set = "SDSP"},

       {
         .name = "W1J",
         .mled_set = "MLED",
         .wled_set = "WLED",
	 .wled_status = "\\WRST",
         .bt_switch = "BLED",
	 .bt_status = "\\BRST",
         .rled_set = "RLED",
         .ledd_set = "SLCM",
         .lcd_switch = xxN_PREFIX "_Q10",
         .lcd_status = xxN_PREFIX "RPIN(0x11)",
         .brightness_set = "SPLV",
         .brightness_get = "GPLV",
         .display_set = "SDSP",
	 .display_get = "\\_SB.PCI0.P0P2.VGA.GETD"},

	{
	 .name = "W1N",
	 .mled_set = "MLED",
	 .wled_set = "WLED",
	 .ledd_set = "SLCM",
	 .lcd_switch = xxN_PREFIX "_Q10",
	 .lcd_status = "\\BKLT",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\ADVG"},

	{
	 .name = "W2V",
	 .mled_set = "MLED",
	 .wled_set = "WLED",
	 .wled_status = "\\WLDF",
	 .bt_switch = "BLED",
	 .bt_status = "\\BLDF",
	 .lcd_switch = xxN_PREFIX "_Q10",
	 .lcd_status = "\\RGPL(0x15,0x01)",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\_SB.PCI0.P0P2.VGA.GETD"},

	{
	 .name = "W3V",
	 .mled_set = "MLED",
	 .wled_set = "WLED",
	 .lcd_switch = xxN_PREFIX "_Q10",
	 .lcd_status = "\\BKLT",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\INFB"},

	{
	 .name = "W3Z",
	 .wled_set = "WLED",
	 .wled_status = "\\WLSF",
	 .bt_switch = "BLED",
	 .bt_status = "\\BLSF",
	 .lcd_switch = xxN_PREFIX "_Q10",
	 .lcd_status = "\\RUOP(0x01)",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\_SB.PCI0.P0P1.VGA.GETD"},

	{
	 .name = "W5A",
	 .bt_switch = "BLED",
	 .wled_set = "WLED",
	 .lcd_switch = xxN_PREFIX "_Q10",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\ADVG"},

        {
	 .name = "W5F",
	 .bt_switch = "BLED",
	 .wled_set = "WLED",
	 .tled_set = "TLED",
	 .lcd_switch = xxN_PREFIX "_Q10",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
         .display_get = "\\ADVG"},

        {
	 .name = "W6A",
	 .bt_switch = "BLED",
	 .wled_set = "WLED",
	 .tled_set = "TLED",
	 .lcd_switch = xxN_PREFIX "_Q10",
	 .lcd_status = "\\RGPL(0x15,0x01)",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
         .display_get = "\\SSTE"},

	{
	 .name = "xxN",
	 .mled_set = "MLED",
/* WLED present, but not controlled by ACPI */
	 .lcd_switch = xxN_PREFIX "_Q10",
	 .lcd_status = "\\BKLT",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\ADVG"},

	{
	 .name = "Z96",
	 .mled_set = "MLED",
	 .wled_set = "WLED",
	 .bt_switch = "BLED",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP",
	 .display_get = "\\_SB.PCI0.VGA.GETD"},

	{
	 .name = "???",
	 .mled_set = "MLED",
	 .wled_set = "WLED",
	 .bt_switch = "BLED",
	 .brightness_set = "SPLV",
	 .brightness_get = "GPLV",
	 .display_set = "SDSP"}
}; /* End methods configuration */

/* procdir we use */
static struct proc_dir_entry *asus_proc_dir;

/*
 * This header is made available to allow proper configuration given model,
 * revision number , ... this info cannot go in struct asus_hotk because it is
 * available before the hotk
 */
static struct acpi_table_header *asus_info;

/* The actual device the driver binds to */
static struct asus_hotk *hotk;

/*
 * The hotkey driver declaration
 */
static int asus_hotk_add(struct acpi_device *device);
static int asus_hotk_remove(struct acpi_device *device, int type);
static struct acpi_driver asus_hotk_driver = {
	.name = ACPI_HOTK_NAME,
	.class = ACPI_HOTK_CLASS,
	.ids = ACPI_HOTK_HID,
	.ops = {
		.add = asus_hotk_add,
		.remove = asus_hotk_remove,
		},
};

#ifdef CONFIG_ACPI_ASUS_NEW_BACKLIGHT
/* The backlight device /sys/class/backlight */
static struct backlight_device *asus_backlight_device;

/*
 * The backlight class declaration
 */
static int read_brightness(struct backlight_device *bd);
static int update_bl_status(struct backlight_device *bd);
static struct backlight_properties asusbl_data = {
	       .owner          = THIS_MODULE,
	       .get_brightness = read_brightness,
	       .update_status  = update_bl_status,
	       .max_brightness = 15,
};
#endif

/*
 * Take a string describing a method,
 * parse it, and store the result into mt
 */
static void parse_method(char *orig, struct method *mt)
{
	char *args = orig;
	uint argc = 0;

	mt->name = strsep(&args, "(");
	mt->input.count = 0;
	mt->input.pointer = mt->params;

	while(args && *args && argc < MT_MAX_ARGS) {
		char *tmp;
		tmp = strsep(&args, ",");

		mt->params[argc].type = ACPI_TYPE_INTEGER;
		mt->params[argc].integer.value = simple_strtoul(tmp, NULL, 0);

		argc++;
	}

	mt->input.count = argc;
}

/*
 * This function evaluates an ACPI method, given an int as parameter, the
 * method is searched within the scope of the handle, can be NULL. The output
 * of the method is written is output, which can also be NULL
 *
 * returns 1 if write is successful, 0 else.
 */
static int write_acpi_int(acpi_handle handle, const char *method, int val,
			  struct acpi_buffer *output)
{
	struct acpi_object_list params;	//list of input parameters (an int here)
	union acpi_object in_obj;	//the only param we use
	acpi_status status;

	params.count = 1;
	params.pointer = &in_obj;
	in_obj.type = ACPI_TYPE_INTEGER;
	in_obj.integer.value = val;

	status = acpi_evaluate_object(handle, (char *)method, &params, output);
	return (status == AE_OK);
}

static int read_acpi_int(acpi_handle handle, const char *method, int *val,
			 struct acpi_object_list *params)
{
	struct acpi_buffer output;
	union acpi_object out_obj;
	acpi_status status;

	output.length = sizeof(out_obj);
	output.pointer = &out_obj;

	status = acpi_evaluate_object(handle, (char *)method, params, &output);
	*val = out_obj.integer.value;
	return (status == AE_OK) && (out_obj.type == ACPI_TYPE_INTEGER);
}

/*
 * We write our info in page, we begin at offset off and cannot write more
 * than count bytes. We set eof to 1 if we handle those 2 values. We return the
 * number of bytes written in page
 */
static int
proc_read_info(char *page, char **start, off_t off, int count, int *eof,
	       void *data)
{
	int len = 0;
	int temp;
	char buf[16];		//enough for all info
	/*
	 * We use the easy way, we don't care of off and count, so we don't set eof
	 * to 1
	 */

	len += sprintf(page, ACPI_HOTK_NAME " " ASUS_ACPI_VERSION "\n");
	len += sprintf(page + len, "Model reference    : %s\n",
		       hotk->methods->name);
	/*
	 * The SFUN method probably allows the original driver to get the list
	 * of features supported by a given model. For now, 0x0100 or 0x0800
	 * bit signifies that the laptop is equipped with a Wi-Fi MiniPCI card.
	 * The significance of others is yet to be found.
	 */
	if (read_acpi_int(hotk->handle, "SFUN", &temp, NULL))
		len +=
		    sprintf(page + len, "SFUN value         : 0x%04x\n", temp);
	/*
	 * Another value for userspace: the ASYM method returns 0x02 for
	 * battery low and 0x04 for battery critical, its readings tend to be
	 * more accurate than those provided by _BST.
	 * Note: since not all the laptops provide this method, errors are
	 * silently ignored.
	 */
	if (read_acpi_int(hotk->handle, "ASYM", &temp, NULL))
		len +=
		    sprintf(page + len, "ASYM value         : 0x%04x\n", temp);
	if (asus_info) {
		snprintf(buf, 16, "%d", asus_info->length);
		len += sprintf(page + len, "DSDT length        : %s\n", buf);
		snprintf(buf, 16, "%d", asus_info->checksum);
		len += sprintf(page + len, "DSDT checksum      : %s\n", buf);
		snprintf(buf, 16, "%d", asus_info->revision);
		len += sprintf(page + len, "DSDT revision      : %s\n", buf);
		snprintf(buf, 7, "%s", asus_info->oem_id);
		len += sprintf(page + len, "OEM id             : %s\n", buf);
		snprintf(buf, 9, "%s", asus_info->oem_table_id);
		len += sprintf(page + len, "OEM table id       : %s\n", buf);
		snprintf(buf, 16, "%x", asus_info->oem_revision);
		len += sprintf(page + len, "OEM revision       : 0x%s\n", buf);
		snprintf(buf, 5, "%s", asus_info->asl_compiler_id);
		len += sprintf(page + len, "ASL comp vendor id : %s\n", buf);
		snprintf(buf, 16, "%x", asus_info->asl_compiler_revision);
		len += sprintf(page + len, "ASL comp revision  : 0x%s\n", buf);
	}

	return len;
}

/*
 * /proc handlers
 * We write our info in page, we begin at offset off and cannot write more
 * than count bytes. We set eof to 1 if we handle those 2 values. We return the
 * number of bytes written in page
 */

/* Generic LED functions */
static int read_led(const char *ledname, int ledmask)
{
	if (ledname) {
		int led_status;

		if (read_acpi_int(NULL, ledname, &led_status, NULL))
			return led_status;
		else
			printk(KERN_WARNING "Asus ACPI: Error reading LED "
			       "status\n");
	}

	return (hotk->status & ledmask) ? 1 : 0;
}

static int parse_arg(const char __user * buf, unsigned long count, int *val)
{
	char s[32];
	if (!count)
		return 0;
	if (count > 31)
		return -EINVAL;
	if (copy_from_user(s, buf, count))
		return -EFAULT;
	s[count] = 0;
	if (sscanf(s, "%i", val) != 1)
		return -EINVAL;
	return count;
}

/* FIXME: kill extraneous args so it can be called independently */
static int
write_led(const char __user * buffer, unsigned long count,
	  char *ledname, int ledmask, int invert)
{
	int rv, value;
	int led_out = 0;

	rv = parse_arg(buffer, count, &value);
	if (rv > 0)
		led_out = value ? 1 : 0;

	hotk->status =
	    (led_out) ? (hotk->status | ledmask) : (hotk->status & ~ledmask);

	if (invert)		/* invert target value */
		led_out = !led_out & 0x1;

	if (!write_acpi_int(hotk->handle, ledname, led_out, NULL))
		printk(KERN_WARNING "Asus ACPI: LED (%s) write failed\n",
		       ledname);

	return rv;
}

/*
 * Proc handlers for MLED
 */
static int
proc_read_mled(char *page, char **start, off_t off, int count, int *eof,
	       void *data)
{
	return sprintf(page, "%d\n",
		       read_led(hotk->methods->mled_status, MLED_ON));
}

static int
proc_write_mled(struct file *file, const char __user * buffer,
		unsigned long count, void *data)
{
	return write_led(buffer, count, hotk->methods->mled_set, MLED_ON, 1);
}

/*
 * Proc handlers for LED display
 */
static int
proc_read_ledd(char *page, char **start, off_t off, int count, int *eof,
	       void *data)
{
	return sprintf(page, "0x%08x\n", hotk->ledd_status);
}

static int
proc_write_ledd(struct file *file, const char __user * buffer,
		unsigned long count, void *data)
{
	int rv, value;

	rv = parse_arg(buffer, count, &value);
	if (rv > 0) {
		if (!write_acpi_int
		    (hotk->handle, hotk->methods->ledd_set, value, NULL))
			printk(KERN_WARNING
			       "Asus ACPI: LED display write failed\n");
		else
			hotk->ledd_status = (u32) value;
	}
	return rv;
}

/*
 * Proc handlers for WLED
 */
static int
proc_read_wled(char *page, char **start, off_t off, int count, int *eof,
	       void *data)
{
	return sprintf(page, "%d\n",
		       read_led(hotk->methods->wled_status, WLED_ON));
}

static int
proc_write_wled(struct file *file, const char __user * buffer,
		unsigned long count, void *data)
{
	return write_led(buffer, count, hotk->methods->wled_set, WLED_ON, 0);
}

/*
 * Proc handlers for Bluetooth
 */
static int
proc_read_bluetooth(char *page, char **start, off_t off, int count, int *eof,
		    void *data)
{
	return sprintf(page, "%d\n", read_led(hotk->methods->bt_status, BT_ON));
}

static int
proc_write_bluetooth(struct file *file, const char __user * buffer,
		     unsigned long count, void *data)
{
	/* Note: bt_switch controls both internal Bluetooth adapter's
	   presence and its LED */
	return write_led(buffer, count, hotk->methods->bt_switch, BT_ON, 0);
}

/*
 * Proc handlers for TLED
 */
static int
proc_read_tled(char *page, char **start, off_t off, int count, int *eof,
	       void *data)
{
	return sprintf(page, "%d\n",
		       read_led(hotk->methods->tled_status, TLED_ON));
}

static int
proc_write_tled(struct file *file, const char __user * buffer,
		unsigned long count, void *data)
{
	return write_led(buffer, count, hotk->methods->tled_set, TLED_ON, 0);
}

/*
 * Proc handlers for RLED
 */
static int
proc_read_rled(char *page, char **start, off_t off, int count, int *eof,
	       void *data)
{
	return sprintf(page, "%d\n",
		       read_led(hotk->methods->rled_status, RLED_ON));
}

static int
proc_write_rled(struct file *file, const char __user * buffer,
		unsigned long count, void *data)
{
	return write_led(buffer, count, hotk->methods->rled_set, RLED_ON, 0);
}

static int get_lcd_state(void)
{
	static struct method mt = { .name = NULL };
	int lcd = 0;

	if(mt.name != hotk->methods->lcd_status)
		parse_method(hotk->methods->lcd_status, &mt);

	if (!read_acpi_int(NULL, mt.name, &lcd, &mt.input))
		printk(KERN_WARNING
		       "Asus ACPI: Error reading LCD status\n");

	if (hotk->model == L2D)
		lcd = ~lcd;
	else if(hotk->model == L3H) /* That's what the AML code does */
		lcd = lcd >> 8;


	return (lcd & 1);
}

static int set_lcd_state(int value)
{
	static struct method mt = { .name = NULL };
	int lcd = 0;
	acpi_status status = 0;

	if(mt.name != hotk->methods->lcd_switch)
		parse_method(hotk->methods->lcd_switch, &mt);

	lcd = value ? 1 : 0;
	if (lcd != get_lcd_state()) {
		if(hotk->model == L3H) /* L3H is special .. */
			status =
				acpi_evaluate_object(hotk->handle,
						     (char *)mt.name,
						     &mt.input, NULL);
		else
			status = acpi_evaluate_object(NULL, (char *)mt.name,
						      &mt.input, NULL);

		if (ACPI_FAILURE(status))
			printk(KERN_WARNING
			       "Asus ACPI: Error switching LCD\n");
	}
	return 0;

}

static int
proc_read_lcd(char *page, char **start, off_t off, int count, int *eof,
	      void *data)
{
	return sprintf(page, "%d\n", get_lcd_state());
}

static int
proc_write_lcd(struct file *file, const char __user * buffer,
	       unsigned long count, void *data)
{
	int rv, value;

	rv = parse_arg(buffer, count, &value);
	if (rv > 0)
		set_lcd_state(value);
	return rv;
}

static int read_brightness(struct backlight_device *bd)
{
	int value;

	if (hotk->methods->brightness_get) {	/* SPLV/GPLV laptop */
		if (!read_acpi_int(hotk->handle, hotk->methods->brightness_get,
				   &value, NULL))
			printk(KERN_WARNING
			       "Asus ACPI: Error reading brightness\n");
	} else if (hotk->methods->brightness_status) {	/* For D1 for example */
		if (!read_acpi_int(NULL, hotk->methods->brightness_status,
				   &value, NULL))
			printk(KERN_WARNING
			       "Asus ACPI: Error reading brightness\n");
	} else			/* No GPLV method */
		value = hotk->brightness;
	return value;
}

/*
 * Change the brightness level
 */
static int set_brightness(struct backlight_device *bd, int value)
{
	acpi_status status = 0;
	int ret = 0;

	value = (0 < value) ? ((15 < value) ? 15 : value) : 0;
	/* 0 <= value <= 15 */

	/* SPLV laptop */
	if (hotk->methods->brightness_set) {
		if (!write_acpi_int(hotk->handle, hotk->methods->brightness_set,
				    value, NULL)) {
			printk(KERN_WARNING
			       "Asus ACPI: Error changing brightness\n");
			ret = -EIO;
			goto out;
		}
	} else {
		/* No SPLV method if we are here, act as appropriate */
		value -= read_brightness(NULL);
		while (value != 0) {
			status = acpi_evaluate_object(NULL, (value > 0) ?
						      hotk->methods->brightness_up :
						      hotk->methods->brightness_down,
						      NULL, NULL);
			(value > 0) ? value-- : value++;
			if (ACPI_FAILURE(status)) {
				printk(KERN_WARNING
				       "Asus ACPI: Error changing brightness\n");
				ret = -EIO;
			}
		}
	}

out:
	return ret;
}

#ifdef CONFIG_ACPI_ASUS_NEW_BACKLIGHT
static int update_bl_status(struct backlight_device *bd)
{
	int value = bd->props->brightness;

	return set_brightness(bd, value);
}
#endif

static int
proc_read_brn(char *page, char **start, off_t off, int count, int *eof,
	      void *data)
{
	return sprintf(page, "%d\n", read_brightness(NULL));
}

static int
proc_write_brn(struct file *file, const char __user * buffer,
	       unsigned long count, void *data)
{
	int rv, value;

	rv = parse_arg(buffer, count, &value);
	if (rv > 0) {
		set_brightness(NULL, value);
	}
	return rv;
}

static void set_display(int value)
{
	/* no sanity check needed for now */
	if (!write_acpi_int(hotk->handle, hotk->methods->display_set,
			    value, NULL))
		printk(KERN_WARNING "Asus ACPI: Error setting display\n");
	return;
}

/*
 * Now, *this* one could be more user-friendly, but so far, no-one has
 * complained. The significance of bits is the same as in proc_write_disp()
 */
static int
proc_read_disp(char *page, char **start, off_t off, int count, int *eof,
	       void *data)
{
	int value = 0;

	if (!read_acpi_int(hotk->handle, hotk->methods->display_get,
			   &value, NULL))
		printk(KERN_WARNING
		       "Asus ACPI: Error reading display status\n");

	value &= 0x0F;		/* needed for some models, shouldn't hurt others */

	return sprintf(page, "%d\n", value);
}

/*
 * Experimental support for display switching. As of now: 1 should activate
 * the LCD output, 2 should do for CRT, 4 for TV-Out and 8 for DVI.
 * Any combination (bitwise) of these will suffice. I never actually tested 4
 * displays hooked up simultaneously, so be warned. See the acpi4asus README
 * for more info.
 */
static int
proc_write_disp(struct file *file, const char __user * buffer,
		unsigned long count, void *data)
{
	int rv, value;

	rv = parse_arg(buffer, count, &value);
	if (rv > 0)
		set_display(value);
	return rv;
}

static void set_light_sens_switch(int value)
{
	if (!write_acpi_int(hotk->handle, hotk->methods->light_sens_switch,
			    value, NULL))
		printk(KERN_WARNING "Asus ACPI: Error setting light sensor "
		       " switch\n");
	hotk->light_switch = value;
}

static int proc_read_lssw(char *page, char **start, off_t off, int count,
			  int *eof, void *data)
{
	return sprintf(page, "%d\n", hotk->light_switch);
}

static int proc_write_lssw(struct file *file, const char __user *buffer,
			   unsigned long count, void *data)
{
	int rv, value;

	rv = parse_arg(buffer, count, &value);
	if (rv > 0)
		set_light_sens_switch(value ? 1 : 0);

	return rv;
}

static void set_light_sens_level(int value)
{
	if (!write_acpi_int(hotk->handle, hotk->methods->light_sens_level,
			    value, NULL))
		printk(KERN_WARNING "Asus ACPI: Error setting light sensor"
		       " level\n");
	hotk->light_level = value;
}

static int
proc_read_lslvl(char *page, char **start, off_t off, int count, int *eof,
		void *data)
{
	return sprintf(page, "%d\n", hotk->light_level);
}

static int
proc_write_lslvl(struct file *file, const char __user *buffer,
		 unsigned long count, void *data)
{
	int rv, value;

	rv = parse_arg(buffer, count, &value);
	if (rv > 0) {
		value = (0 < value) ? ((15 < value) ? 15 : value) : 0;
		/* 0 <= value <= 15 */
		set_light_sens_level(value);
	}

	return rv;
}

typedef int (proc_readfunc) (char *page, char **start, off_t off, int count,
			     int *eof, void *data);
typedef int (proc_writefunc) (struct file * file, const char __user * buffer,
			      unsigned long count, void *data);

static int
asus_proc_add(char *name, proc_writefunc * writefunc,
		     proc_readfunc * readfunc, mode_t mode,
		     struct acpi_device *device)
{
	struct proc_dir_entry *proc =
	    create_proc_entry(name, mode, acpi_device_dir(device));
	if (!proc) {
		printk(KERN_WARNING "  Unable to create %s fs entry\n", name);
		return -1;
	}
	proc->write_proc = writefunc;
	proc->read_proc = readfunc;
	proc->data = acpi_driver_data(device);
	proc->owner = THIS_MODULE;
	proc->uid = 0;
	proc->gid = 0;
	return 0;
}

static int asus_hotk_add_fs(struct acpi_device *device)
{
	struct proc_dir_entry *proc;
	mode_t mode;

	mode = S_IFREG | S_IRUGO | S_IWUGO;

	acpi_device_dir(device) = asus_proc_dir;
	if (!acpi_device_dir(device))
		return -ENODEV;

	proc = create_proc_entry(PROC_INFO, mode, acpi_device_dir(device));
	if (proc) {
		proc->read_proc = proc_read_info;
		proc->data = acpi_driver_data(device);
		proc->owner = THIS_MODULE;
		proc->uid = 0;
		proc->gid = 0;
	} else {
		printk(KERN_WARNING "  Unable to create " PROC_INFO
		       " fs entry\n");
	}

	if (hotk->methods->wled_set) {
		asus_proc_add(PROC_WLED, &proc_write_wled, &proc_read_wled,
			      mode, device);
	}

	if (hotk->methods->ledd_set) {
		asus_proc_add(PROC_LEDD, &proc_write_ledd, &proc_read_ledd,
			      mode, device);
	}

	if (hotk->methods->mled_set) {
		asus_proc_add(PROC_MLED, &proc_write_mled, &proc_read_mled,
			      mode, device);
	}

	if (hotk->methods->rled_set) {
		asus_proc_add(PROC_RLED, &proc_write_rled, &proc_read_rled,
			      mode, device);
	}

	if (hotk->methods->tled_set) {
		asus_proc_add(PROC_TLED, &proc_write_tled, &proc_read_tled,
			      mode, device);
	}

	if (hotk->methods->bt_switch) {
		asus_proc_add(PROC_BT, &proc_write_bluetooth,
			      &proc_read_bluetooth, mode, device);
	}

	/*
	 * We need both read node and write method as LCD switch is also
	 * accessible from keyboard
	 */
	if (hotk->methods->lcd_switch && hotk->methods->lcd_status) {
		asus_proc_add(PROC_LCD, &proc_write_lcd, &proc_read_lcd, mode,
			      device);
	}

	if ((hotk->methods->brightness_up && hotk->methods->brightness_down) ||
	    (hotk->methods->brightness_get && hotk->methods->brightness_set)) {
		asus_proc_add(PROC_BRN, &proc_write_brn, &proc_read_brn, mode,
			      device);
	}

	if (hotk->methods->display_set) {
		asus_proc_add(PROC_DISP, &proc_write_disp, &proc_read_disp,
			      mode, device);
	}

	if (hotk->methods->light_sens_switch) {
		asus_proc_add(PROC_LSSW, &proc_write_lssw, &proc_read_lssw,
			      mode, device);
	}

	if (hotk->methods->light_sens_level) {
		asus_proc_add(PROC_LSLVL, &proc_write_lslvl, &proc_read_lslvl,
			      mode, device);
	}

	return 0;
}

static int asus_hotk_remove_fs(struct acpi_device *device)
{
	if (acpi_device_dir(device)) {
		remove_proc_entry(PROC_INFO, acpi_device_dir(device));
		if (hotk->methods->wled_set)
			remove_proc_entry(PROC_WLED, acpi_device_dir(device));
		if (hotk->methods->mled_set)
			remove_proc_entry(PROC_MLED, acpi_device_dir(device));
		if (hotk->methods->rled_set)
			remove_proc_entry(PROC_RLED, acpi_device_dir(device));
		if (hotk->methods->tled_set)
			remove_proc_entry(PROC_TLED, acpi_device_dir(device));
		if (hotk->methods->ledd_set)
			remove_proc_entry(PROC_LEDD, acpi_device_dir(device));
		if (hotk->methods->bt_switch)
			remove_proc_entry(PROC_BT, acpi_device_dir(device));
		if (hotk->methods->lcd_switch && hotk->methods->lcd_status)
			remove_proc_entry(PROC_LCD, acpi_device_dir(device));
		if ((hotk->methods->brightness_up
		     && hotk->methods->brightness_down)
		    || (hotk->methods->brightness_get
			&& hotk->methods->brightness_set))
			remove_proc_entry(PROC_BRN, acpi_device_dir(device));
		if (hotk->methods->display_set)
			remove_proc_entry(PROC_DISP, acpi_device_dir(device));
		if (hotk->methods->light_sens_switch)
			remove_proc_entry(PROC_LSSW, acpi_device_dir(device));
		if (hotk->methods->light_sens_level)
			remove_proc_entry(PROC_LSLVL, acpi_device_dir(device));
	}
	return 0;
}

static void asus_hotk_notify(acpi_handle handle, u32 event, void *data)
{
	/* TODO Find a better way to handle events count. */
	if (!hotk)
		return;

	if ((event & ~((u32) BR_UP)) < 16) {
		hotk->brightness = (event & ~((u32) BR_UP));
	} else if ((event & ~((u32) BR_DOWN)) < 16) {
		hotk->brightness = (event & ~((u32) BR_DOWN));
	}

	acpi_bus_generate_event(hotk->device, event,
				hotk->event_count[event % 128]++);

	return;
}

/*
 * Match the model string to the list of supported models. Return END_MODEL if
 * no match or model is NULL.
 */
static int asus_model_match(char *model)
{
	if (model == NULL)
		return END_MODEL;

	if (strncmp(model, "L3D", 3) == 0)
		return L3D;
	else if (strncmp(model, "L2E", 3) == 0 ||
		 strncmp(model, "L3H", 3) == 0 || strncmp(model, "L5D", 3) == 0)
		return L3H;
	else if (strncmp(model, "L3", 2) == 0 || strncmp(model, "L2B", 3) == 0)
		return L3C;
	else if (strncmp(model, "L8L", 3) == 0)
		return L8L;
	else if (strncmp(model, "L4R", 3) == 0)
		return L4R;
	else if (strncmp(model, "M6N", 3) == 0 || strncmp(model, "W3N", 3) == 0)
		return M6N;
	else if (strncmp(model, "M6R", 3) == 0 || strncmp(model, "A3G", 3) == 0)
		return M6R;
	else if (strncmp(model, "M6V", 3) == 0 || strncmp(model, "M6A", 3) == 0)
		return M6V;
	else if (strncmp(model, "M7V", 3) == 0)
		return M7V;
	else if (strncmp(model, "M7A", 3) == 0)
		return M7A;
	else if (strncmp(model, "M9V", 3) == 0)
		return M9V;
	else if (strncmp(model, "M2N", 3) == 0 ||
		 strncmp(model, "M3N", 3) == 0 ||
		 strncmp(model, "M5N", 3) == 0 ||
		 strncmp(model, "S1N", 3) == 0 ||
		 strncmp(model, "S5N", 3) == 0)
		return xxN;
	else if (strncmp(model, "Z96", 3) == 0)
		return Z96;
	else if (strncmp(model, "M1", 2) == 0)
		return M1A;
	else if (strncmp(model, "M2", 2) == 0 || strncmp(model, "L4E", 3) == 0)
		return M2E;
	else if (strncmp(model, "L2", 2) == 0)
		return L2D;
	else if (strncmp(model, "F3J", 3) == 0)
		return F3J;
	else if (strncmp(model, "F3T", 3) == 0)
		return F3T;
	else if (strncmp(model, "L8", 2) == 0)
		return S1x;
	else if (strncmp(model, "D1", 2) == 0)
		return D1x;
	else if (strncmp(model, "A1", 2) == 0)
		return A1x;
	else if (strncmp(model, "A2D", 3) == 0)
		return A2D;
	else if (strncmp(model, "A2", 2) == 0)
		return A2x;
	else if (strncmp(model, "A3N", 3) == 0 || strncmp(model, "A3L", 3) == 0)
		return A3N;
	else if (strncmp(model, "A3E", 3) == 0)
		return A3E;
	else if (strncmp(model, "A3F", 3) == 0 || strncmp(model, "A6F", 3) == 0)
		return A3F;
	else if (strncmp(model, "J1", 2) == 0)
		return S2x;
	else if (strncmp(model, "L5", 2) == 0)
		return L5x;
	else if (strncmp(model, "A4G", 3) == 0)
		return A4G;
	else if (strncmp(model, "A4K", 3) == 0 || strncmp(model, "A4D", 3) == 0)
		return A4K;
	else if (strncmp(model, "A4L", 3) == 0)
		return A4L;
	else if (strncmp(model, "A6J", 3) == 0)
		return A6J;
	else if (strncmp(model, "A6R", 3) == 0)
		return A6R;
	else if (strncmp(model, "A6V", 3) == 0 || strncmp(model, "A6Q", 3) == 0)
		return A6V;
	else if (strncmp(model, "A6T", 3) == 0 ||
		 strncmp(model, "A6M", 3) == 0)
		return A6T;
	else if (strncmp(model, "A6B", 3) == 0 ||
		 strncmp(model, "A6K", 3) == 0 ||
		 strncmp(model, "A6U", 3) == 0)
		return A6B;
	else if (strncmp(model, "A7D", 3) == 0)
		return A7D;
	else if (strncmp(model, "A7J", 3) == 0)
		return A7J;
	else if (strncmp(model, "A7T", 3) == 0)
		return A7T;
	else if (strncmp(model, "A8J", 3) == 0 || strncmp(model, "W7J", 3) == 0)
		return A8J;
	else if (strncmp(model, "W1N", 3) == 0)
		return W1N;
	else if (strncmp(model, "W1J", 3) == 0)
		return W1J;
	else if (strncmp(model, "W2V", 3) == 0)
		return W2V;
	else if (strncmp(model, "W3V", 3) == 0)
		return W3V;
	else if (strncmp(model, "W3Z", 3) == 0)
		return W3Z;
	else if (strncmp(model, "W5A", 3) == 0)
		return W5A;
	else if (strncmp(model, "W5F", 3) == 0)
		return W5F;
	else if (strncmp(model, "W6A", 3) == 0)
		return W6A;
	else if (strncmp(model, "VX1", 3) == 0)
		return VX1;
	else if (strncmp(model, "V6V", 3) == 0)
		return V6V;
	else if (strncmp(model, "V6J", 3) == 0)
		return V6J;
	else if (strncmp(model, "S5A", 3) == 0 || strncmp(model, "M5A", 3) == 0)
		return S5A;
	else if (strncmp(model, "U5F", 3) == 0)
		return U5F;
	else
		return END_MODEL;
}

/*
 * This function is used to initialize the hotk with right values. In this
 * method, we can make all the detection we want, and modify the hotk struct
 */
static int asus_hotk_get_info(void)
{
	struct acpi_buffer buffer = { ACPI_ALLOCATE_BUFFER, NULL };
	struct acpi_buffer dsdt = { ACPI_ALLOCATE_BUFFER, NULL };
	union acpi_object *model = NULL;
	int bsts_result;
	char *string = NULL;
	acpi_status status;

	/*
	 * Get DSDT headers early enough to allow for differentiating between
	 * models, but late enough to allow acpi_bus_register_driver() to fail
	 * before doing anything ACPI-specific. Should we encounter a machine,
	 * which needs special handling (i.e. its hotkey device has a different
	 * HID), this bit will be moved. A global variable asus_info contains
	 * the DSDT header.
	 */
	status = acpi_get_table(ACPI_TABLE_DSDT, 1, &dsdt);
	if (ACPI_FAILURE(status))
		printk(KERN_WARNING "  Couldn't get the DSDT table header\n");
	else
		asus_info = dsdt.pointer;

	/* We have to write 0 on init this far for all ASUS models */
	if (!write_acpi_int(hotk->handle, "INIT", 0, &buffer)) {
		printk(KERN_ERR "  Hotkey initialization failed\n");
		return -ENODEV;
	}

	/* This needs to be called for some laptops to init properly */
	if (!read_acpi_int(hotk->handle, "BSTS", &bsts_result, NULL))
		printk(KERN_WARNING "  Error calling BSTS\n");
	else if (bsts_result)
		printk(KERN_NOTICE "  BSTS called, 0x%02x returned\n",
		       bsts_result);

	/*
	 * Try to match the object returned by INIT to the specific model.
	 * Handle every possible object (or the lack of thereof) the DSDT
	 * writers might throw at us. When in trouble, we pass NULL to
	 * asus_model_match() and try something completely different.
	 */
	if (buffer.pointer) {
		model = buffer.pointer;
		switch (model->type) {
		case ACPI_TYPE_STRING:
			string = model->string.pointer;
			break;
		case ACPI_TYPE_BUFFER:
			string = model->buffer.pointer;
			break;
		}
	}
	hotk->model = asus_model_match(string);
	if (hotk->model == END_MODEL) {	/* match failed */
		kfree(model);
		if (asus_info &&
		    strncmp(asus_info->oem_table_id, "ODEM", 4) == 0) {
			hotk->model = P30;
			printk(KERN_NOTICE
			       "  Samsung P30 detected, supported\n");
		} else if(unsupported) {
			printk(KERN_NOTICE "  unsupported model %s, trying "
			       "default values\n", string);
			printk(KERN_NOTICE
			       "  send /proc/acpi/dsdt to the developers\n");
			hotk->model = DEF;

		} else {
			printk(KERN_NOTICE "  unsupported model %s\n", string);
			printk(KERN_NOTICE
			       "  send /proc/acpi/dsdt to the developers\n");
			printk(KERN_NOTICE
			       "  you can use the \"unsupported=1\" "
			       "parameter to have a minimal support\n");
			return -ENODEV;
		}
		hotk->methods = &model_conf[hotk->model];
		return AE_OK;
	}
	hotk->methods = &model_conf[hotk->model];
	printk(KERN_NOTICE "  %s model detected, supported\n", string);

	/* Sort of per-model blacklist */
	if (strncmp(string, "L2B", 3) == 0)
		hotk->methods->lcd_status = NULL;
	/* L2B is similar enough to L3C to use its settings, with this only
	   exception */
	else if (strncmp(string, "A3G", 3) == 0)
		hotk->methods->lcd_status = "\\BLFG";
	/* A3G is like M6R */
	else if (strncmp(string, "S5N", 3) == 0 ||
		 strncmp(string, "M5N", 3) == 0)
		hotk->methods->mled_set = NULL;
	/* S5N, M5N and W3N have no MLED */
	else if (strncmp(string, "L5D", 3) == 0)
		hotk->methods->wled_set = NULL;
	/* L5D's WLED is not controlled by ACPI */
	else if (strncmp(string, "M2N", 3) == 0 ||
		 strncmp(string, "S1N", 3) == 0)
		hotk->methods->wled_set = "WLED";
	/* M2N and S1N have a usable WLED */
	else if (strncmp(string, "W7J", 3) == 0)
		hotk->methods->mled_set = NULL;
	/* W7J is like A8J, but without MLED */
	else if (asus_info) {
		if (strncmp(asus_info->oem_table_id, "L1", 2) == 0)
			hotk->methods->mled_status = NULL;
		/* S1300A reports L84F, but L1400B too, account for that */
	}

	kfree(model);

	return AE_OK;
}

static int asus_hotk_check(void)
{
	int result = 0;

	result = acpi_bus_get_status(hotk->device);
	if (result)
		return result;

	if (hotk->device->status.present) {
		result = asus_hotk_get_info();
	} else {
		printk(KERN_ERR "  Hotkey device not present, aborting\n");
		return -EINVAL;
	}

	return result;
}

static int asus_hotk_found;

static int asus_hotk_add(struct acpi_device *device)
{
	acpi_status status = AE_OK;
	int result;

	if (!device)
		return -EINVAL;

	printk(KERN_NOTICE "Asus Laptop ACPI Extras version %s\n",
	       ASUS_ACPI_VERSION);

	hotk = kmalloc(sizeof(struct asus_hotk), GFP_KERNEL);
	if (!hotk)
		return -ENOMEM;
	memset(hotk, 0, sizeof(struct asus_hotk));

	hotk->handle = device->handle;
	strcpy(acpi_device_name(device), ACPI_HOTK_DEVICE_NAME);
	strcpy(acpi_device_class(device), ACPI_HOTK_CLASS);
	acpi_driver_data(device) = hotk;
	hotk->device = device;

	result = asus_hotk_check();
	if (result)
		goto end;

	result = asus_hotk_add_fs(device);
	if (result)
		goto end;

	/*
	 * We install the handler, it will receive the hotk in parameter, so, we
	 * could add other data to the hotk struct
	 */
	status = acpi_install_notify_handler(hotk->handle, ACPI_SYSTEM_NOTIFY,
					     asus_hotk_notify, hotk);
	if (ACPI_FAILURE(status))
		printk(KERN_ERR "  Error installing notify handler\n");

	/* For laptops without GPLV: init the hotk->brightness value */
	if ((!hotk->methods->brightness_get)
	    && (!hotk->methods->brightness_status)
	    && (hotk->methods->brightness_up && hotk->methods->brightness_down)) {
		status =
		    acpi_evaluate_object(NULL, hotk->methods->brightness_down,
					 NULL, NULL);
		if (ACPI_FAILURE(status))
			printk(KERN_WARNING "  Error changing brightness\n");
		else {
			status =
			    acpi_evaluate_object(NULL,
						 hotk->methods->brightness_up,
						 NULL, NULL);
			if (ACPI_FAILURE(status))
				printk(KERN_WARNING "  Strange, error changing"
				       " brightness\n");
		}
	}

	asus_hotk_found = 1;

	/* LED display is off by default */
	hotk->ledd_status = 0xFFF;

	/* Set initial values of light sensor and level */
	hotk->light_switch = 1;/* Default to light sensor disabled */
	hotk->light_level = 0;/* level 5 for sensor sensitivity */

	if (hotk->methods->light_sens_switch)
		set_light_sens_switch(hotk->light_switch);

	if (hotk->methods->light_sens_level)
		set_light_sens_level(hotk->light_level);

      end:
	if (result) {
		kfree(hotk);
	}

	return result;
}

static int asus_hotk_remove(struct acpi_device *device, int type)
{
	acpi_status status = 0;

	if (!device || !acpi_driver_data(device))
		return -EINVAL;

	status = acpi_remove_notify_handler(hotk->handle, ACPI_SYSTEM_NOTIFY,
					    asus_hotk_notify);
	if (ACPI_FAILURE(status))
		printk(KERN_ERR "Asus ACPI: Error removing notify handler\n");

	asus_hotk_remove_fs(device);

	kfree(hotk);

	return 0;
}

static void __exit asus_acpi_exit(void)
{
#ifdef CONFIG_ACPI_ASUS_NEW_BACKLIGHT
	if(asus_backlight_device) {
# if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,19)
		down(&asus_backlight_device->sem);
		asus_backlight_device->props->update_status = NULL;
		up(&asus_backlight_device->sem);
# endif
		backlight_device_unregister(asus_backlight_device);
	}
#endif
	acpi_bus_unregister_driver(&asus_hotk_driver);
	remove_proc_entry(PROC_ASUS, acpi_root_dir);

	kfree(asus_info);

	return;
}

static int __init asus_acpi_init(void)
{
#ifdef CONFIG_ACPI_ASUS_NEW_BACKLIGHT
	struct backlight_device *bd;
#endif
	int result;

	if (acpi_disabled)
		return -ENODEV;

	if (!acpi_specific_hotkey_enabled) {
		printk(KERN_ERR "Using generic hotkey driver\n");
		return -ENODEV;
	}
	asus_proc_dir = proc_mkdir(PROC_ASUS, acpi_root_dir);
	if (!asus_proc_dir) {
		printk(KERN_ERR "Asus ACPI: Unable to create /proc entry\n");
		return -ENODEV;
	}
	asus_proc_dir->owner = THIS_MODULE;

	result = acpi_bus_register_driver(&asus_hotk_driver);
	if (result < 0) {
		remove_proc_entry(PROC_ASUS, acpi_root_dir);
		return result;
	}

	/*
	 * This is a bit of a kludge.  We only want this module loaded
	 * for ASUS systems, but there's currently no way to probe the
	 * ACPI namespace for ASUS HIDs.  So we just return failure if
	 * we didn't find one, which will cause the module to be
	 * unloaded.
	 */
	if (!asus_hotk_found) {
		acpi_bus_unregister_driver(&asus_hotk_driver);
		remove_proc_entry(PROC_ASUS, acpi_root_dir);
		return -ENODEV;
	}

#ifdef CONFIG_ACPI_ASUS_NEW_BACKLIGHT
	bd = backlight_device_register ("asus", NULL, &asusbl_data);

	if (IS_ERR (bd)) {
		printk(KERN_ERR "Could not register asus backlight device\n");
		asus_backlight_device = NULL;
		asus_acpi_exit();
	}

	asus_backlight_device = bd;

	down(&bd->sem);
	bd->props->brightness = read_brightness(NULL);
	up(&bd->sem);
#endif

	return 0;
}

module_init(asus_acpi_init);
module_exit(asus_acpi_exit);
