/*
  Copyright 2005 Denis Vlasenko

  This file is part of prism54usb.

  prism54usb is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  prism54usb is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with prism54usb; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

*/

#ifndef _HAVE_ISLSL_LOG_H_
#define _HAVE_ISLSL_LOG_H_

#include <linux/types.h>
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/jiffies.h>

#define ISL_DEBUG 2

/***********************************************************************
** Logging: guidelines
**
** - Use printk() for messages which are to be always logged.
**   Supply either 'islsm:' or '<devname>:' prefix so that user
**   can figure out who's speaking among other kernel chatter.
**   islsm: is for general issues (e.g. "islsm: no firmware image!")
**   while <devname>: is related to a particular device
**   (think about multi-card setup). Double check that message
**   is not confusing to the average user.
**
** - use printk KERN_xxx level only if message is not a WARNING
**   but is INFO, ERR etc. In general, we mostly not using these
**
** - Use islog() for messages which may be omitted (and they
**   _will_ be omitted in non-debug builds). Note that
**   message levels may be disabled at compile-time selectively,
**   thus select them wisely. Example: L_DEBUG is the lowest
**   (most likely to be compiled out) -> use for less important stuff.
**
** - Do not print important stuff with islog(), or else people
**   will never build non-debug driver.
*/

enum {
	/* generic debug */
	L_DEBUG = (ISL_DEBUG > 1) * 0x0001,
	/* function tracing -- through debugfs & relayfs */
	L_FUNC = (ISL_DEBUG > 0) * 0x0002,
	/* IRQ data, use with care */
	L_IRQ = (ISL_DEBUG > 0) * 0x0004,
	/* UART data -- through debugfs & relayfs */
	L_DATA =  (ISL_DEBUG > 1) * 0x0008,
	/* device IOCTLs */
	L_IOCTL = (ISL_DEBUG > 0) * 0x0010,
	/* raw protocol islsm frames
	   input (R) and output (T) -- through debugfs & relayfs */
	L_BUFR = (ISL_DEBUG > 1) * 0x0020,
	L_BUFT = (ISL_DEBUG > 1) * 0x0040,
	/* firmware & PDA parsing infos */
	L_FW = (ISL_DEBUG > 0) * 0x0080,
	L_PDA = (ISL_DEBUG > 0) * 0x0100,
	/* allocator debug */
	L_ALLOC = (ISL_DEBUG > 1) * 0x0200,
	/* frame interpretation, input, output -- through debugfs & relayfs */
	L_SM_INPUT  = (ISL_DEBUG > 0) * 0x0400,
	L_SM_OUTPUT = (ISL_DEBUG > 0) * 0x0800,
	L_ANY = 0xffff,
};

#if ISL_DEBUG
extern unsigned int     isl_debug;
#else
enum { isl_debug = 0 };
#endif

#if ISL_DEBUG > 1

void                    isl_fn_enter(const char *funcname);
void                    isl_fn_exit(const char *funcname);
void                    isl_fn_exit_v(const char *funcname, int v);

#define FN_ENTER \
	do { \
		if (unlikely(isl_debug & L_FUNC)) { \
			isl_fn_enter(__func__); \
		} \
	} while (0)

#define FN_EXIT1(v) \
	do { \
		if (unlikely(isl_debug & L_FUNC)) { \
			isl_fn_exit_v(__func__, v); \
		} \
	} while (0)
#define FN_EXIT0 \
	do { \
		if (unlikely(isl_debug & L_FUNC)) { \
			isl_fn_exit(__func__); \
		} \
	} while (0)

#else

#define FN_ENTER
#define FN_EXIT1(v)
#define FN_EXIT0

#endif				/* ISL_DEBUG > 1 */

#if ISL_DEBUG

#define islog(chan, args...) \
	do { \
		if (isl_debug & (chan)) \
			printk(args); \
	} while (0)

void                    isl_dump_bytes(const void *data, int num);

#else				/* Non-debug build: */

#define islog(chan, args...)
#define isl_dump_bytes(data, num)

#endif				/* ISL_DEBUG */

void                    isl_print_mac(const char *head, const u8 *mac,
				      const char *tail);

/* Optimized out to nothing in non-debug build */
static inline void
islog_mac(int level, const char *head, const u8 *mac, const char *tail)
{
	if (isl_debug & level) {
		isl_print_mac(head, mac, tail);
	}
}

/* Useful to get __FILE__ without full leading path */
static inline const char *sanitize_str(const char *s)
{
	const char             *t = strrchr(s, '/');
	if (t)
		return t + 1;
	return s;
}

/* raw output data debug */
static inline void
islsm_txdata_debug(const unsigned int queue,
		   const void *_data, const int len)
{
	if (isl_debug & L_BUFT) {
		unsigned long int time = jiffies;
		printk(KERN_CRIT "[OUT +%-6ld] %02x\n", time, queue);
		isl_dump_bytes(_data, len);
	}
}

/* raw input data debug */
static inline void
islsm_rxdata_debug(const unsigned int queue,
		   const void *data, const int len)
{
	if (isl_debug & L_BUFR) {
		unsigned long time = jiffies;
		printk(KERN_CRIT "[IN +%-6ld] %02x\n", time, queue);
		isl_dump_bytes(data, len);
	}
}

#endif				/* _HAVE_ISLSL_LOG_H_ */
