/*
 * Connection tracking state replication for Netfilter
 *
 * (C) 2002-2004, KOVACS Krisztian <hidden@sch.bme.hu>
 * (C) 2003-2004, Harald Welte <laforge@netfilter.org>
 * (C) 2004, BalaBit IT Ltd. <www.balabit.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <linux/config.h>
#include <linux/version.h>

#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/prefetch.h>
#include <linux/vmalloc.h>
#include <linux/spinlock.h>

#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/skbuff.h>
#include <net/sock.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <linux/inetdevice.h>
#include <linux/igmp.h>
#include <linux/ip.h>		/* for sizeof(struct iphdr) */
#include <linux/udp.h>		/* for sizeof(struct udphdr) */

/* no locking at the moment... */
#define ASSERT_READ_LOCK(x)
#define ASSERT_WRITE_LOCK(x)
#include <linux/netfilter_ipv4/listhelp.h>

#include <linux/netfilter_ipv4/ct_sync.h>
#include <linux/netfilter_ipv4/ct_sync_sock.h>
#include <linux/netfilter_ipv4/ct_sync_proto.h>
#include <linux/netfilter_ipv4/cts_buff.h>

#define MTU			1500
#define CT_SEND_DATASIZE	(MTU-sizeof(struct iphdr) 		\
				 -sizeof(struct udphdr) 		\
				 -sizeof(struct ct_sync_pkthdr))
#define CT_SYNC_SEND_BACKLOG	32

struct cts_proto_sock {
	struct csb_ring ring;
	struct socket *socket;
	wait_queue_head_t *wait;
};

struct cts_protoh {
	/* tx and rx sockets */
	struct cts_proto_sock send;
	struct cts_proto_sock recv;
	/* node state */
	atomic_t state;
	/* callback function to call on state transitions */
	void (*transition_callback)(struct cts_protoh *,
				    enum cts_proto_state,
				    enum cts_proto_state,
				    void *);
	void *transition_callback_data;
	/* timestamp of last initsync request */
	unsigned long last_initsync;
	/* recovery state of the node */
	struct {
		atomic_t active;
		__u16 seq;
		int diff;
	} recovery_request;
};

static const char *cts_proto_statenames[0x15] = {
	[CT_SYNC_PSTATE_NONE] 		= "NONE",
	[CT_SYNC_PSTATE_SLAVE_INIT] 	= "SLAVE_INIT",
	[CT_SYNC_PSTATE_SLAVE_SYNSENT]	= "SLAVE_SYNSENT",
	[CT_SYNC_PSTATE_SLAVE_SYNRECV]	= "SLAVE_SYNRECV",
	[CT_SYNC_PSTATE_SLAVE_RUNNING]	= "SLAVE_RUNNING",
	[CT_SYNC_PSTATE_MASTER_INIT]	= "MASTER_INIT",
	[CT_SYNC_PSTATE_MASTER_ADVSENT]	= "MASTER_ADVSENT",
	[CT_SYNC_PSTATE_MASTER_RUNNING]	= "MASTER_RUNNING",
};

static inline void
set_state(struct cts_protoh *cph, unsigned int state)
{
	unsigned int oldstate = atomic_read(&cph->state);
	atomic_set(&cph->state, state);
	CT_SYNC_INFO("changed state: %s -> %s\n",
		     cts_proto_statenames[oldstate],
		     cts_proto_statenames[state]);
	if (oldstate != state
	    && cph->transition_callback != NULL)
		cph->transition_callback(cph, oldstate, state,
					 cph->transition_callback_data);
}

int 
cts_proto_is_master(struct cts_protoh *cph)
{
	return (atomic_read(&cph->state) >= CT_SYNC_PSTATE_MASTER_INIT);
}

int
cts_proto_get_state(struct cts_protoh *cph)
{
	return atomic_read(&cph->state);
}

/***********************************************************************
 * SYSCTL functions
 ***********************************************************************/

/* upper limit is one minute */
#define MSG_AGE_LIMIT 60000
static unsigned long max_msg_age = 2 * HZ;

unsigned int
cts_proto_get_max_age(void)
{
	return jiffies_to_msecs(max_msg_age);
}

int
cts_proto_set_max_age(unsigned int msecs)
{
	if (msecs > MSG_AGE_LIMIT)
		return -EINVAL;

	max_msg_age = msecs_to_jiffies(msecs);

	return 0;
}

/***********************************************************************
 * CTS_BUFF / CSB_RING util functions
 ***********************************************************************/
static inline void
csb_hdr_fill(struct cts_protoh *cph, struct ct_sync_pkthdr *hdr, 
	     __u8 type, __u16 seq)
{
	hdr->version = 2;
	hdr->pkttype = type;
	hdr->pktseq = htons(seq);
	hdr->count = 0;
}

static inline void
init_csb(struct cts_buff *csb)
{
	csb->tail = (char *) &csb->pkt.data[0];
	csb->parsed = csb->tail;
}
	
/* allocate a new cts_buff */
static struct cts_buff *
alloc_csb(unsigned int size, int gfp_mask)
{
	struct cts_buff *csb = kmalloc(sizeof(*csb)+size, gfp_mask);

	CT_SYNC_ENTER();

	if (!csb) {
		CT_SYNC_LEAVE();
		return NULL;
	}

	memset(csb, 0, sizeof(*csb)+size);
	csb->end = (char *) &csb->pkt.data[0] + size;
	atomic_set(&csb->use, 0);

	init_csb(csb);
	
	CT_SYNC_LEAVE();

	return csb;
}

static void
kfree_csb(struct cts_buff *csb)
{
	CT_SYNC_ENTER();

	if (unlikely(atomic_read(&csb->use))) {
		CT_SYNC_ERR("freeing csb that is still used\n");
		CT_SYNC_LEAVE();
		return;
	}

	kfree(csb);

	CT_SYNC_LEAVE();
	return;
}

/* Duplicate a cts_buff and return a pointer to the new structure */
static struct cts_buff *
dup_csb(const struct cts_buff *csb, int gfp_mask)
{
	struct cts_buff *ncsb;
	size_t size;

	CT_SYNC_ENTER();

	CT_SYNC_ASSERT(csb);

	size = (size_t)(csb->end - (char *) csb + 1);
	ncsb = kmalloc(size, gfp_mask);
	if (ncsb == NULL) {
		CT_SYNC_LEAVE();
		return NULL;
	}

	memset(ncsb, 0, sizeof(*ncsb));
	ncsb->tail = (char *)&ncsb->pkt + csb_len(csb);
	ncsb->end = (char *)&ncsb->pkt + csb_size(csb);
	ncsb->parsed = (char *)&ncsb->pkt.data[0];
	memcpy(&ncsb->pkt, &csb->pkt, csb_len(csb));

	CT_SYNC_LEAVE();

	return ncsb;
}

/* Initialize an already-allocated csb_ring with 'num' number of cts_buffs,
 * each of 'size' size.
 *
 * A few words about locking of the rings. The receiver ring's locking is simple,
 * since that is accessed only by our receiver kthread. For the sender ring, we have
 * to use spin_{lock,unlock}_bh(), since that ring is accessed by our conntrack
 * notifier functions as well.
 */
static int
csb_ring_init(struct csb_ring *csr, unsigned int num, unsigned int size,
		unsigned int backlog)
{
	unsigned int i;

	CT_SYNC_ENTER();

	spin_lock_init(&csr->lock);

	for (i = 0; i < num; i++) {
		struct cts_buff *d = alloc_csb(size, GFP_KERNEL);
		if (unlikely(!d)) {
			CT_SYNC_ERR("ENOMEM while allocating send_desc #%u\n",
					i+1);
			CT_SYNC_LEAVE();
			/* FIXME: free allocated buffers */
			return -ENOMEM;
		}
		
		d->ring = csr;
#if (CTSYNC_DEBUG >= 4)
		d->ringnum = i;
#endif
		if (unlikely(i == 0)) {
			INIT_LIST_HEAD(&d->list);
			csr->cur = &d->list;
		} else
			list_add_tail(&d->list, csr->cur);
	}
	/* cur now points to first allocated item */

	/* initialize sent */
	csr->sent = csr->cur->prev;
	csr->backlog = csr->sent;

	/* the only 'allocated' buffer is the next one */
	csr->alloc = csr->cur->next;

	/* initialize backlog.  we should actually start with 0 and have it
	 * grow while sending packets. later */ 
	for (i = 0; i < backlog; i++)
		csr->backlog = csr->backlog->prev;

	CT_SYNC_LEAVE();
	return 0;
}

static void
csb_ring_destroy(struct csb_ring *csr)
{
	struct list_head *cur, *next;	

	CT_SYNC_ENTER();

	spin_lock_bh(&csr->lock);

	list_for_each_safe(cur, next, csr->cur) {
		kfree_csb((struct cts_buff *)cur);
	}
	kfree_csb((struct cts_buff *)csr->cur);

	spin_unlock_bh(&csr->lock);

	CT_SYNC_LEAVE();
}

/* advance to next cts_buffer in ring. allocate, if no free cts_buff
 *
 * WARNING: in spite of filling the packet header, this function does not
 * increment the ring's sequence number, since it is used by the receiver code
 * as well. The _want_enqueue() function increments the sequence number
 * after advancing to the next buffer, while the receiver code handles the seqno
 * of the ring specially (it is the sequence number of the last successfully
 * received packet). */
static struct cts_buff *
__csb_ring_advance_alloc(struct cts_protoh *cph,
			 struct csb_ring *csr)
{ 
	struct cts_buff *csb;

	CT_SYNC_ENTER();

	if (unlikely(csr->alloc->next == csr->backlog)) {
		/* full. need to allocate one more */
		CT_SYNC_DEBUG("csr full, need to allocate more\n");
		csb = alloc_csb(CT_SEND_DATASIZE, GFP_ATOMIC);
		if (!csb) {
			CT_SYNC_ERR("unable to allocate cts_buff\n");
			CT_SYNC_LEAVE();
			return NULL;
		}
		csb->ring = csr;
		/* add new entry just before backlog */
		list_add(&csb->list, csr->backlog->prev);
	}

	/* make sure we really can advance alloc ptr */
	CT_SYNC_ASSERT(csr->alloc->next != csr->backlog);
	CT_SYNC_ASSERT(csr->alloc->next != csr->sent);

	/* advance alloc pointer */
	csr->alloc = csr->alloc->next;

	csb = (struct cts_buff *) csr->alloc;
	csb->firstuse = 0UL;
	/* fill pkt header */
	csb_hdr_fill(cph, &csb->pkt.hdr, CT_SYNC_PKT_SYNC, 0);

	CT_SYNC_LEAVE();

	return csb;
}

/***********************************************************************
 * CONTROL MESSAGES
 ***********************************************************************/

/* we have just been told to become master; tell other nodes */
int
cts_proto_master_announce(struct cts_protoh *cph)
{
	ssize_t len;
	struct {
		struct ct_sync_pkthdr hdr;
		struct ct_sync_master_announce ma;
	} pkt;

	CT_SYNC_ENTER();

	CT_SYNC_DEBUG("sending master_announce with seqno=%u\n",
		      cph->send.ring.seqno);

	memset(&pkt, 0, sizeof(pkt));
	csb_hdr_fill(cph, &pkt.hdr, CT_SYNC_PKT_MASTER_ANNOUNCE, 
		     cph->send.ring.seqno);

	/* FIXME: wraps every five days */
	pkt.ma.upsince = htonl(jiffies);
	pkt.ma.flags = 0;

	len = cts_sock_send(cph->send.socket, (char *)&pkt, sizeof(pkt));
	if (unlikely(len < sizeof(pkt))) {
		CT_SYNC_ERR("Send failed, returned %d.\n", len);
		CT_SYNC_LEAVE();
		return -1;
	}

	CT_SYNC_LEAVE();

	return 0;
}

/* externally triggered: we become master.  called from _main.c */
int
cts_proto_become_master(struct cts_protoh *cph)
{
	CT_SYNC_ENTER();

	CTS_STAT_INC(slave2master);

	set_state(cph, CT_SYNC_PSTATE_MASTER_INIT);

	if (unlikely(cts_proto_master_announce(cph) < 0)) {
		CT_SYNC_ERR("unable to become master\n");
		set_state(cph, CT_SYNC_PSTATE_NONE);
		CT_SYNC_LEAVE();
		return -1;
	}
	set_state(cph, CT_SYNC_PSTATE_MASTER_ADVSENT);

	set_state(cph, CT_SYNC_PSTATE_MASTER_RUNNING);

	CT_SYNC_LEAVE();

	return 0;
}

/* externally triggered: we become slave */
int
cts_proto_become_slave(struct cts_protoh *cph)
{
	CT_SYNC_ENTER();

	set_state(cph, CT_SYNC_PSTATE_SLAVE_INIT);

	CT_SYNC_LEAVE();

	return 0;
}

/* externally triggered: we're signalling a full re-sync */
int
cts_proto_initsync_follows(struct cts_protoh *cph)
{
	ssize_t len;
	struct {
		struct ct_sync_pkthdr hdr;
	} pkt;

	CT_SYNC_ENTER();

	memset(&pkt, 0, sizeof(pkt));
	csb_hdr_fill(cph, &pkt.hdr, CT_SYNC_PKT_INITSYNC_FOLLOWS, 0);

	len = cts_sock_send(cph->send.socket, (unsigned char *)&pkt, sizeof(pkt));
	if (unlikely(len < sizeof(pkt))) {
		CT_SYNC_ERR("Send failed, returned %d.\n", len);
		CT_SYNC_LEAVE();
		return -1;
	}

	CT_SYNC_LEAVE();

	return 0;
}

/* send an initsync request */
static int
cts_proto_send_initsync_request(struct cts_protoh *cph)
{
	ssize_t len;
	struct {
		struct ct_sync_pkthdr hdr;
		struct ct_sync_slave_initsync is;
	} pkt;

	CT_SYNC_ENTER();

	CTS_STAT_INC(initsyncs);

	CT_SYNC_DEBUG("sending INITSYNC request\n");

	memset(&pkt, 0, sizeof(pkt));
	csb_hdr_fill(cph, &pkt.hdr, CT_SYNC_PKT_SLAVE_INITSYNC, 0);

	len = cts_sock_send(cph->send.socket, (unsigned char*)&pkt, sizeof(pkt));
	if (unlikely(len < sizeof(pkt))) {
		CT_SYNC_ERR("Send failed, only sent %d of %u.\n",
			    len, sizeof(pkt));
		CT_SYNC_LEAVE();
		return -1;
	}

	CT_SYNC_LEAVE();

	return 0;
}

/* externally triggered: transmit an initsync request if necessary:
 *   - we're in state CT_SYNC_PSTATE_SLAVE_INIT 
 *   - we're in state CT_SYNC_PSTATE_SLAVE_SYNSENT or _SYNRECV, but at least
 *     five seconds have elapsed since the last request
 */
int
cts_proto_initsync_request(struct cts_protoh *cph)
{
	unsigned int state = atomic_read(&cph->state);

	if ((state == CT_SYNC_PSTATE_SLAVE_INIT)
	    || ((state == CT_SYNC_PSTATE_SLAVE_SYNSENT
		 || state == CT_SYNC_PSTATE_SLAVE_SYNRECV)
		&& jiffies >= cph->last_initsync + 5 * HZ)) {
		cts_proto_send_initsync_request(cph);
		cph->last_initsync = jiffies;
		if (state == CT_SYNC_PSTATE_SLAVE_INIT)
			set_state(cph, CT_SYNC_PSTATE_SLAVE_SYNSENT);
	}

	return 0;
}

/***********************************************************************
 * SENDER
 ***********************************************************************/

/* advance ring's current buffer pointer if possible
 * must be called with the ring's spinlock held
 * returns: 1 if there was at least one unused buffer
 * 	    0 otherwise
 */
static int
cts_proto_advance_cur(struct csb_ring *csr)
{
	int need_wakeup = 0;

	CT_SYNC_ENTER();

	while (!csb_used((struct cts_buff *)csr->cur)) {
		/* make sure we stop before alloc; we wake up
		 * sender only if there is at least one fully
		 * filled packet, that is, alloc was advanced
		 * by _want_enqueue(), which means that 
		 * cur != alloc->prev */
		if (csr->cur == csr->alloc->prev)
			break;
		/* advance and wake up sender */
		csr->cur = csr->cur->next;
		need_wakeup = 1;
	}

	CT_SYNC_LEAVE();

	return need_wakeup;
}

/* decrement usage count and advance csb->cur if we hit zero */
void
csb_use_dec(struct cts_protoh *cph, struct cts_buff *csb)
{
	CT_SYNC_ENTER();
	DUMP_CTS_BUFF(csb);

	if (likely(atomic_dec_and_test(&csb->use))) {
		if (unlikely(!csb->ring)) {
			CT_SYNC_ERR("decrementing usage count on csb without ring?\n");
			CT_SYNC_LEAVE();
			return;
		}

		spin_lock_bh(&csb->ring->lock);

		/* nobody using us, can advance ringbuf */
		if (cts_proto_advance_cur(csb->ring) == 1) {
			//CT_SYNC_DEBUG("trying to wake up sender thread\n");
			wake_up(cph->send.wait);
		}

		spin_unlock_bh(&csb->ring->lock);
	}

	CT_SYNC_LEAVE();
}

/* caller wants to send the current packet even if it is not full.
 * returns 0 on success, -1 otherwise */
int
cts_proto_advance_timeout(struct cts_protoh *cph)
{
	struct cts_buff *tmp;
	struct csb_ring *csr;
	
	CT_SYNC_ENTER();
	
	csr = &cph->send.ring;

	spin_lock_bh(&csr->lock);

	tmp = (struct cts_buff *)csr->alloc;

	/* check if
	 *   1. the buffer is currently unused
	 *   2. we're not right after init
	 *   3. the current message is old enough
	 */
	if (!csb_used(tmp)
	    && (tmp != (struct cts_buff *)csr->cur)
	    && tmp->firstuse && (jiffies - tmp->firstuse > max_msg_age)) {
		tmp = __csb_ring_advance_alloc(cph, csr);
		if (unlikely(tmp == NULL)) {
			spin_unlock_bh(&csr->lock);
			CT_SYNC_ERR("unable to advance ringbuffer\n");
			CT_SYNC_LEAVE();
			return -1;
		}

		/* as we have successfully allocated a new buffer for next
		 * messages, advance current pointer to indicate to the sender
		 * thread that there are messages waiting to be sent */
		cts_proto_advance_cur(csr);

		csr_print(csr);
	}

	spin_unlock_bh(&csr->lock);

	CT_SYNC_LEAVE();

	return 0;
}

/* caller wants to enqueue `size' bytes. we have to return pointer to csb that
 * is capable of that many bytes */
void *
cts_proto_want_enqueue(struct cts_protoh *cph,
		       struct cts_buff **pcsb, unsigned int size)
{
	void *retbuf;
	struct csb_ring *csr = &cph->send.ring;
	struct cts_buff *tmp;

	CT_SYNC_ENTER();

	spin_lock_bh(&csr->lock);

	csr_print(csr);

	/* we can only check last 'alloc' buffer, not the other used
	 * ones between 'cur' and 'alloc' - otherwise we'd violate ordering
	 * constraints */
	tmp = (struct cts_buff *)csr->alloc;
	/* we have to advance the alloc pointer if any of the following is true:
	 *   1. special case after startup: csr->alloc == csr->cur
	 *   2. the current buffer's first use timestamp is old enough
	 *   3. not enough space, need to advance/allocate new csb
	 */
	if (tmp == (struct cts_buff *)csr->cur
	    || (tmp->firstuse && (jiffies - tmp->firstuse >= max_msg_age))
	    || !(retbuf = csb_put(tmp, size))) {
		tmp = __csb_ring_advance_alloc(cph, csr);
		if (unlikely(!tmp)) {
			spin_unlock_bh(&csr->lock);
			CT_SYNC_ERR("unable to advance ringbuffer\n");
			CT_SYNC_LEAVE();
			return NULL;
		}
	 	retbuf = csb_put(tmp, size);
	}

	CT_SYNC_ASSERT(retbuf);

	csb_use_inc(tmp);

	CT_SYNC_DEBUG2("using buffer %p: ", tmp);
	DUMP_CTS_BUFF(tmp);

	/* increment msg count in packet */
	tmp->pkt.hdr.count++;
	/* initialize first use timestamp if this is the first message */
	if (tmp->pkt.hdr.count == 1)
		tmp->firstuse = jiffies;

	/* return csb */
	*pcsb = tmp;

	csr_print(csr);

	spin_unlock_bh(&csr->lock);

	CT_SYNC_LEAVE();
	return retbuf;
}

int
cts_proto_send_pending(struct cts_protoh *cph)
{
	int ret = 1;

	CT_SYNC_ENTER();

	spin_lock_bh(&cph->send.ring.lock);
	if (cph->send.ring.sent == cph->send.ring.cur) 
		ret = 0;
	spin_unlock_bh(&cph->send.ring.lock);

	CT_SYNC_LEAVE();

	return ret;
}

int
cts_proto_send_dequeue(struct cts_protoh *cph)
{
	struct cts_buff *csb;
	ssize_t len;

	CT_SYNC_ENTER();

	spin_lock_bh(&cph->send.ring.lock);

	if (cph->send.ring.sent == cph->send.ring.cur) {
		spin_unlock(&cph->send.ring.lock);
		/* no pending entries to be sent */
		CT_SYNC_DEBUG2("no pending entries to be sent, returning\n");
		CT_SYNC_LEAVE();
		return -2;
	}
	csb = (struct cts_buff *) cph->send.ring.sent->next;

	spin_unlock_bh(&cph->send.ring.lock);

	if (likely(cts_proto_is_master(cph))) {
		/* advance_alloc() fills the packet header, but does not 
		 * increment our sequence number, so we have to do it here */
		csb->pkt.hdr.pktseq = htons(cph->send.ring.seqno++);
		/* send packet */
		CT_SYNC_DEBUG2("sending csb=%p, len=%u\n", csb, csb_len(csb));
		len = cts_sock_send(cph->send.socket, (char *) &csb->pkt, 
				    csb_len(csb));
		if (len < csb_len(csb)) {
			CT_SYNC_ERR("short send: %d\n", len);
			CT_SYNC_LEAVE();
			return -1;
		}
	} else {
		CT_SYNC_INFO("want to send packet while being slave?\n");
	}

	spin_lock_bh(&cph->send.ring.lock);

	/* reset csb data for csb's out of backlog */
	init_csb((struct cts_buff *)cph->send.ring.backlog);
	/* advance 'sent' and 'backlog' pointer by one */
	cph->send.ring.sent = cph->send.ring.sent->next;
	cph->send.ring.backlog = cph->send.ring.backlog->next;

	spin_unlock_bh(&cph->send.ring.lock);

	CT_SYNC_LEAVE();

	return 0;
}

/* Compare the sequence numbers cseq and seq.
 *
 * Return value:
 *   <0 if seq is larger than cseq
 *    0 if equal
 *   >0 if seq is smaller than cseq
 */
static inline int
cts_seqno_cmp(const u16 cseq, const u16 seq)
{
	int diff = cseq - seq + 0x18fff;

	return (diff & 0xffff) - 0x8fff;
}

/* Recover lost packets
 *
 * This must be called from process context!
 *
 * If the seqno of the first backlog buffer (csr->backlog) is larger than the
 * seqno of the first missing packet (seqno in the request plus one) we have no
 * chance to recover, initiate a full re-synchronization. Otherwise resend
 * missing packets.
 *
 * Because cts_sock_send() may sleep, we have to do this in two steps:
 *   - we have to duplicate each csb to be sent with the spinlock of the ring
 *     held
 *   - walk the list of our duplicated buffers and send each csb
 *
 * Returns:
 *  -1 if full re-sync is required
 *   0 otherwise
 */
static int
cts_proto_recover(struct cts_protoh *cph, const u16 seq) 
{ 
	struct csb_ring *csr = &cph->send.ring;
	struct list_head *i, *t, q;
	u16 bseq;

	CT_SYNC_ENTER();

	INIT_LIST_HEAD(&q);

	spin_lock_bh(&csr->lock);

	/* find first valid packet in backlog */
	for (i = csr->backlog; i != csr->sent; i = i->next) {
		struct cts_buff *csb = (struct cts_buff *) i;

		if (ntohs(csb->pkt.hdr.pktseq) != 0)
			break;
	}

	bseq = ntohs(((struct cts_buff *)i)->pkt.hdr.pktseq);

	/* check seqno of csr->backlog */
	if (cts_seqno_cmp(bseq, (u16)((seq + 1) & 0xffff)) > 0) {
		/* give up and request initsync */
		CT_SYNC_DEBUG("recovery impossible: backlog begins at %d\n", bseq);
		spin_unlock_bh(&csr->lock);
		CT_SYNC_LEAVE();
		return -1;
	}

	/* collect buffers */
	for (i = csr->sent; i != csr->backlog->prev; i = i->prev) {
		struct cts_buff *csb = (struct cts_buff *) i;
		struct cts_buff *ncsb;
		u16 cseq = ntohs(csb->pkt.hdr.pktseq);

		/* check if we have to resend this packet */
		if (cts_seqno_cmp(cseq, seq) <= 0)
			continue;

		CT_SYNC_DEBUG("duplicating pkt %d\n", cseq);

		/* copy this csb and add to our queue */
		ncsb = dup_csb(csb, GFP_ATOMIC);
		if (ncsb == NULL) {
			CT_SYNC_DEBUG("failed to dup csb\n");
			continue;
		}
		ncsb->pkt.hdr.flags |= CT_SYNC_PKT_F_RECOVER;
		list_add(&ncsb->list, &q);
	}

	spin_unlock_bh(&csr->lock);

	list_for_each_safe(i, t, &q) {
		struct cts_buff *csb = (struct cts_buff *) i;

		CTS_STAT_INC(tx.pkts_recover);

		/* resend packet */
		ssize_t len = cts_sock_send(cph->send.socket, (char *) &csb->pkt, 
					csb_len(csb));
		if (unlikely(len < csb_len(csb))) {
			CT_SYNC_ERR("error during retransmit: %d\n", len);
		}

		CT_SYNC_DEBUG("re-sent seq %d\n", ntohs(csb->pkt.hdr.pktseq));

		/* free csb */
		list_del(i);
		kfree_csb(csb);
	}

	CT_SYNC_LEAVE();

	return 0;
}

/***********************************************************************
 * RECEIVER
 ***********************************************************************/

/* 
 * (re)initialize recovery structure
 */
static void
cts_proto_recovery_init(struct cts_protoh *cph)
{
	atomic_set(&cph->recovery_request.active, 0);
	cph->recovery_request.seq = 0;
	cph->recovery_request.diff = 0;
}

unsigned int cts_proto_recovery_threshold = 8;

/*
 * request recovery, last seen packet is seq
 */
static int
cts_proto_request_recovery(struct cts_protoh *cph, u16 seq, int diff)
{
	struct ct_sync_pkthdr hdr;
	ssize_t len;

	CT_SYNC_ENTER();

	/* check if we already have a pending request */
	if (atomic_read(&cph->recovery_request.active)) {
		/* if there are at least recovery_threshold lost messages
		 * since the last request, repeat; return immediately otherwise */
		if (diff < cph->recovery_request.diff + cts_proto_recovery_threshold) {
			CT_SYNC_LEAVE();
			return 0;
		}
	}

	atomic_set(&cph->recovery_request.active, 1);
	cph->recovery_request.seq = seq;
	cph->recovery_request.diff = diff;

	memset(&hdr, 0, sizeof(hdr));
	csb_hdr_fill(cph, &hdr, CT_SYNC_PKT_NACK, seq);

	len = cts_sock_send(cph->send.socket, (char *)&hdr, 
			    sizeof(struct ct_sync_pkthdr));
	if (unlikely(len < sizeof(struct ct_sync_pkthdr))) {
		CT_SYNC_ERR("Send failed, returned %d.\n", len);
		CT_SYNC_LEAVE();
		return -1;
	}

	CT_SYNC_LEAVE();

	return 0;
}

/* 
 * Check if a pending recovery request is fulfilled when 'seq' is received
 *
 * We consider a pending request fulfilled, if seq >= request.seq + request.diff.
 * This means that we have to successfully receive the packet which caused
 * the request to be sent.
 *
 * Returns:
 *   0 if there were no pending recovery requests
 *   1 if the request was cancelled
 *  -1 if the request is still pending
 */
static int
cts_proto_cancel_recovery(struct cts_protoh *cph, u16 seq)
{
	int ret = 0, diff;
	
	CT_SYNC_ENTER();
	
	if (atomic_read(&cph->recovery_request.active)) {
		diff = cts_seqno_cmp(seq, cph->recovery_request.seq);

		CT_SYNC_DEBUG("request seq %d diff %d received %d\n",
			      cph->recovery_request.seq,
			      cph->recovery_request.diff, seq);

		if (diff >= cph->recovery_request.diff) {
			/* cancel request */
			atomic_set(&cph->recovery_request.active, 0);
			CT_SYNC_DEBUG("cancelled request\n");
		}
	}

	CT_SYNC_LEAVE();

	return ret;
}

/* Receive packet, process packet header info. That is:
 *  - For invalid packets, discard the packet.
 *    Return value: CTS_PROTO_RCV_DISCARDED
 *  - For SYNC packets, check sequence number, and request retransmission
 *    or enqueue the packet in the receive buffer.
 *    Return value: CTS_PROTO_RCV_SYNC
 *  - For NACK packets, initiate re-send of packets, and discard the NACK.
 *    Return value: CTS_PROTO_RCV_NACK
 *  - For MASTER_ANNOUNCE, update the receive ring's sequence number,
 *    and discard the packet.
 *    Return value: CTS_PROTO_RCV_MASTERANN
 *  - For INITSYNC, notify the caller that we received an INITSYNC request,
 *    and discard the packet.
 *    Return value: CTS_PROTO_RCV_INITSYNC
 *
 * In case of unrecoverable errors, the function returns CTS_PROTO_RCV_ERROR.
 */
int
cts_proto_recv_pkt(struct cts_protoh *cph)
{
	struct csb_ring *csr = &cph->recv.ring;
	struct cts_buff *csb;

	ssize_t len;
	int ret = CTS_PROTO_RCV_DISCARDED;

	CT_SYNC_ENTER();

	/* FIXME: do we need a spinlock here? Only called from kthread,
	 * which is serialized. Seems that the receive ring is accessed
	 * only from the receiver thread, so locking shouldn't be a problem.
	 */
	spin_lock(&csr->lock);

	/* advance */
	csb =__csb_ring_advance_alloc(cph, csr);
	if (unlikely(!csb)) {
		ret = CTS_PROTO_RCV_ERROR;
		CT_SYNC_LEAVE();
		goto done_unlock;
	}

	/* receive packet */
	len = cts_sock_recv(cph->recv.socket, (char *)&csb->pkt, csb_size(csb));
	if (unlikely(len < sizeof(struct ct_sync_pkthdr))) {
		CT_SYNC_ERR("short packet, %d < %u\n", len,
				sizeof(struct ct_sync_pkthdr));
		goto done_ignore;
	}
	
	/* tail points to last byte of received data */
	csb->tail = (char *)&csb->pkt + len;
	csb->parsed = (char *)&csb->pkt.data[0];
	
	CT_SYNC_DEBUG("filled csb: %p: ", csb);
	DUMP_CTS_BUFF(csb);

	if (unlikely(csb->pkt.hdr.version != 2)) {
		CT_SYNC_ERR("invalid protocol version %u\n", 
			    csb->pkt.hdr.version);
		goto done_ignore;
	}

	if (csb->pkt.hdr.flags & CT_SYNC_PKT_F_RECOVER)
		CTS_STAT_INC(rx.pkts_recover);

	switch (csb->pkt.hdr.pkttype) {
		__u16 recover_seq;

	case CT_SYNC_PKT_SYNC:

		switch (atomic_read(&cph->state)) {
		case CT_SYNC_PSTATE_SLAVE_INIT:
			/* do we really want to accept sync pkts before we
			 * requested them? For now, fallthrough */
		case CT_SYNC_PSTATE_SLAVE_SYNSENT:
			/* set initial seqno */
			csr->seqno = ntohs(csb->pkt.hdr.pktseq) - 1;
			printk(KERN_DEBUG "setting initial seqno to %u\n",
				csr->seqno);
			set_state(cph, CT_SYNC_PSTATE_SLAVE_SYNRECV);
			break;
		case CT_SYNC_PSTATE_SLAVE_SYNRECV:
		case CT_SYNC_PSTATE_SLAVE_RUNNING:
			break;
		default:
			CT_SYNC_ERR("rcvd sync pkt, but we're not slave!\n");
			goto done_ignore;
			break;
		}

		if (unlikely(ntohs(csb->pkt.hdr.pktseq) != 
					(__u16)((csr->seqno + 1) & 0xffff))) {
			int sdiff;

			CTS_STAT_INC(rx.pkts_ooseq);

			CT_SYNC_INFO("received out-of-sequence sync pkt "
				     "(rcv:%u, exp:%u)\n",
				     ntohs(csb->pkt.hdr.pktseq), csr->seqno + 1);
			sdiff = cts_seqno_cmp(ntohs(csb->pkt.hdr.pktseq), 
					      csr->seqno);

			if (sdiff > 0) {
				/* we have lost some packages, try to recover */
				cts_proto_request_recovery(cph, csr->seqno, 
							   sdiff);
			} else {
				/* we've already seen this, let's drop it */
				CT_SYNC_DEBUG("dropping duplicate, diff %d\n", 
					      sdiff);
			}
			goto done_ignore;
		} else {
			csr->seqno++;
			ret = CTS_PROTO_RCV_SYNC;

			cts_proto_cancel_recovery(cph, csr->seqno);
		}

		break;

	case CT_SYNC_PKT_NACK:
		/* NACK's are processed inline and immediately overwritten
		 * before the next packet is received */
		if (atomic_read(&cph->state) != CT_SYNC_PSTATE_MASTER_RUNNING) {
			CT_SYNC_ERR("rcvd NACK pkt, but we're not master\n");
			goto done_ignore;
		}
		recover_seq = ntohs(csb->pkt.hdr.pktseq);
		CT_SYNC_DEBUG("requested recovery of seq %u\n", recover_seq);
		/* need to go back one in order to overwrite this one */
		csr->alloc = csr->alloc->prev;
		/* free spinlock early to avoid deadlock with send lock */
		spin_unlock(&csr->lock);
		/* try to recover and request full resync if failed */
		if (cts_proto_recover(cph, recover_seq) < 0)
			ret = CTS_PROTO_RCV_INITSYNC;
		else
			ret = CTS_PROTO_RCV_NACK;
		goto done_nounlock;
		break;

	case CT_SYNC_PKT_MASTER_ANNOUNCE:
		switch (atomic_read(&cph->state)) {
		case CT_SYNC_PSTATE_MASTER_INIT:
		case CT_SYNC_PSTATE_MASTER_ADVSENT:
		case CT_SYNC_PSTATE_MASTER_RUNNING:
			CT_SYNC_INFO("new master, degrading to client\n");
			CTS_STAT_INC(master2slave_auto);
			set_state(cph, CT_SYNC_PSTATE_SLAVE_SYNSENT);
			/* fall thru */
		default:
			/* we have a new master, we have to update our receiver's
			 * sequence number */
			CT_SYNC_INFO("received master announcement, "
				     "setting initial seqno to %u\n", 
				     csr->seqno);
			/* csr->seqno is last received seqno, thus we expect
			 * csr->seqno+1 in next SYNC packet. Sender tells us
			 * the sequence number it is about to send with the
			 * next packet, so we need to subtract 1 */
			csr->seqno = ntohs(csb->pkt.hdr.pktseq) - 1;
			break;
		}
		ret = CTS_PROTO_RCV_MASTERANN;
		goto done_ignore;
		break;
		
	case CT_SYNC_PKT_SLAVE_INITSYNC:
		switch (atomic_read(&cph->state)) {
		case CT_SYNC_PSTATE_MASTER_INIT:
			CT_SYNC_ERR("received initsync before announcing ourselves, discarding\n");
			break;
		case CT_SYNC_PSTATE_MASTER_ADVSENT:
		case CT_SYNC_PSTATE_MASTER_RUNNING:
			CT_SYNC_INFO("received initsync request\n");
			ret = CTS_PROTO_RCV_INITSYNC;
			break;
		default:
			CT_SYNC_INFO("received initsync request while being slave; staying cool\n");
			break;
		}
		goto done_ignore;
		break;

	case CT_SYNC_PKT_INITSYNC_FOLLOWS:
		switch (atomic_read(&cph->state)) {
		case CT_SYNC_PSTATE_MASTER_INIT:
		case CT_SYNC_PSTATE_MASTER_ADVSENT:
		case CT_SYNC_PSTATE_MASTER_RUNNING:
			CT_SYNC_DEBUG("received initsync announcement while being master, discarding\n");
			break;
		case CT_SYNC_PSTATE_SLAVE_INIT:
		case CT_SYNC_PSTATE_SLAVE_SYNSENT:
			CT_SYNC_DEBUG("received initsync announcement, good\n");
			break;
		case CT_SYNC_PSTATE_SLAVE_SYNRECV:
		case CT_SYNC_PSTATE_SLAVE_RUNNING:
			CT_SYNC_DEBUG("received initsync announcement while being slave, reinit\n");
			cts_proto_recovery_init(cph);
			set_state(cph, CT_SYNC_PSTATE_SLAVE_SYNSENT);
			break;
		}
		goto done_ignore;
		break;

	default:
		CT_SYNC_ERR("rcvd unknown packet type %u\n", 
			    csb->pkt.hdr.pkttype);
		goto done_ignore;
		break;
	}

	/* we did not ignore the packet, update the cur pointer of the ring */
	csr->cur = &csb->list;
done_unlock:
	spin_unlock(&csr->lock);
done_nounlock:
	CT_SYNC_LEAVE();
	return ret;

done_ignore:
	/* need to go back one in order to overwrite this one */
	csr->alloc = csr->alloc->prev;
	spin_unlock(&csr->lock);
	CT_SYNC_LEAVE();
	return ret;
}

int
cts_proto_recv_pending(struct cts_protoh *cph)
{
	return (!skb_queue_empty(&cph->recv.socket->sk->sk_receive_queue));
}

/* get a pointer to the next to-be-processed message */
struct ct_sync_msghdr *
cts_proto_recvmsg(struct cts_protoh *cph, struct ct_sync_pkthdr **hdr)
{
	struct csb_ring *csr = &cph->recv.ring;
	struct cts_buff *csb = (struct cts_buff *) csr->sent;
	struct ct_sync_msghdr *msghdr;
	int advance_pkt = 0;
	unsigned int msglen;

	CT_SYNC_ENTER();

	/* FIXME: clean up locking, should the receive ring's spinlock
	 * used at all? */
	//spin_lock(&csr->lock);

	csr_print(csr);

	while (1) {
		DUMP_CTS_BUFF(csb);
		if (advance_pkt) {
			/* advance to next packet if requested */
			init_csb(csb);
			if (csr->sent == csr->cur) {
				/* no more to-be-received packets */
				CT_SYNC_DEBUG("no more pkts\n");
				//spin_unlock(&csr->lock);
				CT_SYNC_LEAVE();
				return NULL;
			}
			/* advance 'sent' and backlog by one */
			CT_SYNC_DEBUG("advancing to next packet\n");
			csr->sent = csr->sent->next;
			csr->backlog = csr->backlog->next;
			csb = (struct cts_buff *) csr->sent;
			csr_print(csr);
			DUMP_CTS_BUFF(csb);
			advance_pkt = 0;
		}

		if (csb->parsed >= csb->tail) {
			CT_SYNC_DEBUG("end of packet reached\n");
			advance_pkt = 1;
			continue;
		}

		if ((csb->tail - csb->parsed) < sizeof(*msghdr)) {
			/* no full msghdr ?!?,advance */
			CT_SYNC_DEBUG("no full msghdr\n");
			advance_pkt = 1;
			continue;
		}

		msghdr = (struct ct_sync_msghdr *) csb->parsed;
		msglen = (__u16)ntohs(msghdr->len) + sizeof(*msghdr);

		if ((csb->tail - csb->parsed) < msglen) {
			/* packet is likely to be corrupted or truncated, advance */
			CT_SYNC_DEBUG("truncated message\n");
			advance_pkt = 1;
			continue;
		}

		csb->parsed += msglen;
		*hdr = &csb->pkt.hdr;
		break;
	}

	CT_SYNC_ASSERT(msghdr);

	csr_print(csr);
	//spin_unlock(&csr->lock);

	/* upon receiving a message with the INITSYNC flag set, we
	 * transition from SLAVE_INIT to SLAVE_SYNRECV */
	if (unlikely(msghdr->type == CT_SYNC_MSG_UPDATE
	             && msghdr->flags & CTS_UPD_F_INITSYNC
		     && atomic_read(&cph->state) == CT_SYNC_PSTATE_SLAVE_INIT))
		set_state(cph, CT_SYNC_PSTATE_SLAVE_SYNRECV);

	if (unlikely(msghdr->type == CT_SYNC_MSG_UPDATE
		     && msghdr->flags & CTS_UPD_F_INITSYNC_DONE
		     && atomic_read(&cph->state) == CT_SYNC_PSTATE_SLAVE_SYNRECV))
		set_state(cph, CT_SYNC_PSTATE_SLAVE_RUNNING);

	CT_SYNC_LEAVE();

	return msghdr;
}

/* data ready callback: wake up rcv thread */
static void
cts_proto_data_ready(struct sock *sk, int count)
{
	struct cts_protoh *cph;

	CT_SYNC_ENTER();

	cph = (struct cts_protoh *)sk->sk_user_data;
	if (cph && cph->recv.wait && waitqueue_active(cph->recv.wait))
		wake_up_interruptible(cph->recv.wait);

	if (sk->sk_sleep && waitqueue_active(sk->sk_sleep))
		wake_up_interruptible(sk->sk_sleep);

	CT_SYNC_LEAVE();
}

/* initialize protocol */
struct cts_protoh *
cts_proto_init(char *devname, 
	       const struct sockaddr_in *addr, 
	       void (*transition_callback)(struct cts_protoh*,
					   enum cts_proto_state,
					   enum cts_proto_state,
					   void*),
	       void *transition_callback_data,
	       wait_queue_head_t *send_wait,
	       wait_queue_head_t *rcv_wait)
{
	int ret;
	struct cts_protoh *cph;

	CT_SYNC_ENTER();

	/* allocate new protohandle */
	cph = kmalloc(sizeof(*cph), GFP_KERNEL);
	if (!cph) {
		ret = -ENOMEM;
		goto error;
	}

	memset(cph, 0, sizeof(*cph));

	/* don't call set_state since cph->state is not initialized yet */
	atomic_set(&cph->state, CT_SYNC_PSTATE_NONE);
	cph->transition_callback = transition_callback;
	cph->transition_callback_data = transition_callback_data;

	/* init recovery structure */
	cts_proto_recovery_init(cph);

	/* send queue */
	ret = csb_ring_init(&cph->send.ring, 50, CT_SEND_DATASIZE, 
			CT_SYNC_SEND_BACKLOG);
	if (ret < 0)
		goto error_cph;
	cph->send.ring.seqno = jiffies;
	cph->send.wait = send_wait;
	/* alloc points to cur, special case for startup to defer
	   header initialization to advance_alloc() */
	cph->send.ring.alloc = cph->send.ring.cur = cph->send.ring.sent;
	csr_print(&cph->send.ring);

	/* receive buffer, no backlog */
	ret = csb_ring_init(&cph->recv.ring, 10, CT_SEND_DATASIZE, 1);
	if (ret < 0)
		goto error_csr;
	cph->recv.ring.seqno = 0;
	/* make it look like all received packets have been fully processed */
	cph->recv.ring.sent = cph->recv.ring.cur = cph->recv.ring.alloc;
	cph->recv.ring.backlog = cph->recv.ring.sent->prev;
	csr_print(&cph->recv.ring);
	cph->recv.wait = rcv_wait;

	/* create sockets */
	if ((cph->send.socket = cts_sock_server_init(devname, addr)) == NULL) {
		CT_SYNC_ERR("Failed to create send socket.\n");
		goto error_crr;
	}
	if ((cph->recv.socket = cts_sock_client_init(devname, addr)) == NULL) {
		CT_SYNC_ERR("Failed to create rcv socket.\n");
		goto error_ssock;
	}

	/* setup sk_data_ready callback of receive socket */
	cph->recv.socket->sk->sk_user_data = cph;
	cph->recv.socket->sk->sk_data_ready = cts_proto_data_ready;

	if (ret < 0)
		goto error_rsock;

	CT_SYNC_LEAVE();

	return cph;

error_rsock:
	sock_release(cph->recv.socket);
error_ssock:
	sock_release(cph->send.socket);
error_crr:
	csb_ring_destroy(&cph->recv.ring);
error_csr:
	csb_ring_destroy(&cph->send.ring);
error_cph:
	kfree(cph);
error:
	CT_SYNC_LEAVE();

	return NULL;
}

void
cts_proto_cleanup(struct cts_protoh *cph)
{
	CT_SYNC_ENTER();

	sock_release(cph->recv.socket);
	sock_release(cph->send.socket);
	csb_ring_destroy(&cph->recv.ring);
	csb_ring_destroy(&cph->send.ring);
	kfree(cph);

	CT_SYNC_LEAVE();
}
