/*************************************************** */
/* Rule Set Based Access Control                     */
/* Implementation of AEF LSM hooks                   */
/* Author and (c) 2003: Amon Ott <ao@rsbac.org>      */
/*                                                   */
/* Last modified: 05/Sep/2003                        */
/*************************************************** */

#include <linux/types.h>
#include <linux/version.h>
#include <linux/security.h>
#include <linux/ptrace.h>
#include <linux/fs.h>
#include <linux/mount.h>
#include <linux/namei.h>
#include <net/sock.h>
#include <rsbac/types.h>
#include <rsbac/adf.h>
#include <rsbac/aci.h>

static int rsbac_aef_ptrace(struct task_struct * parent, struct task_struct * child)
  {
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_ptrace(): calling ADF\n");
#endif
    rsbac_target_id.process = child->pid;
    rsbac_attribute_value.trace_request = PTRACE_TRACEME;
    if (!rsbac_adf_request(R_TRACE,
                           parent->pid,
                           T_PROCESS,
                           rsbac_target_id,
                           A_trace_request,
                           rsbac_attribute_value))
      return -EPERM;
    else
      return 0;
  }

static int rsbac_aef_statfs(struct super_block * sb)
  {
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_statfs(): calling ADF\n");
#endif
    rsbac_target_id.dev.type = D_block;
    rsbac_target_id.dev.id = sb->s_dev;
    rsbac_attribute_value.dummy = 0;
    if (!rsbac_adf_request(R_GET_STATUS_DATA,
                           current->pid,
                           T_DEV,
                           rsbac_target_id,
                           A_none,
                           rsbac_attribute_value))
      return -EPERM;
    else
      return 0;
  }

static int rsbac_aef_inode_follow_link(struct dentry *dentry, struct nameidata *nd)
  {
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_inode_follow_link(): calling ADF\n");
#endif
    rsbac_target_id.symlink.device = dentry->d_sb->s_dev;
    rsbac_target_id.symlink.inode  = dentry->d_inode->i_ino;
    rsbac_target_id.symlink.dentry_p = dentry;
    rsbac_attribute_value.dummy = 0;
    if (!rsbac_adf_request(R_SEARCH,
                           current->pid,
                           T_SYMLINK,
                           rsbac_target_id,
                           A_none,
                           rsbac_attribute_value))
      return -EPERM;
    else
      return 0;
  }

static int rsbac_aef_inode_create(struct inode *dir,
	                          struct dentry *dentry,
	                          int mode)
  {
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_inode_create(): calling ADF\n");
#endif
    rsbac_target_id.dir.device = dir->i_sb->s_dev;
    rsbac_target_id.dir.inode  = dir->i_ino;
    rsbac_target_id.dir.dentry_p = dentry->d_parent;
    rsbac_attribute_value.create_data.target = T_FILE;
    rsbac_attribute_value.create_data.dentry_p = dentry;
    rsbac_attribute_value.create_data.mode = mode;
    rsbac_attribute_value.create_data.device = RSBAC_ZERO_DEV;
    if (!rsbac_adf_request(R_CREATE,
                           current->pid,
                           T_DIR,
                           rsbac_target_id,
                           A_create_data,
                           rsbac_attribute_value))
      return -EPERM;
    else
      return 0;
  }

static void rsbac_aef_inode_post_create(struct inode *dir,
	                                struct dentry *dentry,
                                        int mode)
  {
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_target_id_t       rsbac_new_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_inode_post_create(): notifying ADF\n");
#endif
    rsbac_target_id.dir.device = dir->i_sb->s_dev;
    rsbac_target_id.dir.inode  = dir->i_ino;
    rsbac_target_id.dir.dentry_p = dentry->d_parent;
    rsbac_attribute_value.create_data.target = T_FILE;
    rsbac_attribute_value.create_data.dentry_p = dentry;
    rsbac_attribute_value.create_data.mode = mode;
    rsbac_attribute_value.create_data.device = RSBAC_ZERO_DEV;
    rsbac_new_target_id.file.device = dentry->d_sb->s_dev;
    rsbac_new_target_id.file.inode  = dentry->d_inode->i_ino;
    rsbac_new_target_id.file.dentry_p = dentry;
    if (rsbac_adf_set_attr(R_CREATE,
                           current->pid,
                           T_DIR,
                           rsbac_target_id,
                           T_FILE,
                           rsbac_new_target_id,
                           A_create_data,
                           rsbac_attribute_value))
      {
        printk(KERN_WARNING
               "rsbac_aef_inode_post_create() [vfs_create() [open_namei() [filp_open() [do_open() [sys_open()]]]]]: rsbac_adf_set_attr() returned error");
      }
  }

static int rsbac_aef_inode_mknod(struct inode *dir, struct dentry *dentry,
	                         int mode, dev_t dev)
  {
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_inode_mknod(): calling ADF\n");
#endif
    rsbac_target_id.dir.device = dir->i_sb->s_dev;
    rsbac_target_id.dir.inode  = dir->i_ino;
    rsbac_target_id.dir.dentry_p = dentry->d_parent;
    rsbac_attribute_value.create_data.target = T_FILE;
    rsbac_attribute_value.create_data.dentry_p = dentry;
    rsbac_attribute_value.create_data.mode = mode;
    rsbac_attribute_value.create_data.device = dev;
    if (!rsbac_adf_request(R_CREATE,
                           current->pid,
                           T_DIR,
                           rsbac_target_id,
                           A_create_data,
                           rsbac_attribute_value))
      return -EPERM;
    else
      return 0;
  }

static void rsbac_aef_inode_post_mknod(struct inode *dir, struct dentry *dentry,
                                       int mode, dev_t dev)
  {
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_target_id_t       rsbac_new_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_inode_post_mknod(): notifying ADF\n");
#endif
    rsbac_target_id.dir.device = dir->i_sb->s_dev;
    rsbac_target_id.dir.inode  = dir->i_ino;
    rsbac_target_id.dir.dentry_p = dentry->d_parent;
    rsbac_attribute_value.create_data.target = T_FILE;
    rsbac_attribute_value.create_data.dentry_p = dentry;
    rsbac_attribute_value.create_data.mode = mode;
    rsbac_attribute_value.create_data.device = dev;
    rsbac_new_target_id.file.device = dentry->d_sb->s_dev;
    rsbac_new_target_id.file.inode  = dentry->d_inode->i_ino;
    rsbac_new_target_id.file.dentry_p = dentry;
    if (rsbac_adf_set_attr(R_CREATE,
                           current->pid,
                           T_DIR,
                           rsbac_target_id,
                           T_FILE,
                           rsbac_new_target_id,
                           A_create_data,
                           rsbac_attribute_value))
      {
        printk(KERN_WARNING
               "rsbac_aef_inode_post_mknod() [vfs_mknod() [sys_mknod()]: rsbac_adf_set_attr() returned error");
      }
  }

static int rsbac_aef_inode_mkdir(struct inode *dir,
	                          struct dentry *dentry,
	                          int mode)
  {
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_inode_mkdir(): calling ADF\n");
#endif
    rsbac_target_id.dir.device = dir->i_sb->s_dev;
    rsbac_target_id.dir.inode  = dir->i_ino;
    rsbac_target_id.dir.dentry_p = dentry->d_parent;
    rsbac_attribute_value.create_data.target = T_FILE;
    rsbac_attribute_value.create_data.dentry_p = dentry;
    rsbac_attribute_value.create_data.mode = mode;
    rsbac_attribute_value.create_data.device = RSBAC_ZERO_DEV;
    if (!rsbac_adf_request(R_CREATE,
                           current->pid,
                           T_DIR,
                           rsbac_target_id,
                           A_create_data,
                           rsbac_attribute_value))
      return -EPERM;
    else
      return 0;
  }

static void rsbac_aef_inode_post_mkdir(struct inode *dir,
	                                struct dentry *dentry,
                                        int mode)
  {
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_target_id_t       rsbac_new_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_inode_post_mkdir(): notifying ADF\n");
#endif
    rsbac_target_id.dir.device = dir->i_sb->s_dev;
    rsbac_target_id.dir.inode  = dir->i_ino;
    rsbac_target_id.dir.dentry_p = dentry->d_parent;
    rsbac_attribute_value.create_data.target = T_FILE;
    rsbac_attribute_value.create_data.dentry_p = dentry;
    rsbac_attribute_value.create_data.mode = mode;
    rsbac_attribute_value.create_data.device = RSBAC_ZERO_DEV;
    rsbac_new_target_id.file.device = dentry->d_sb->s_dev;
    rsbac_new_target_id.file.inode  = dentry->d_inode->i_ino;
    rsbac_new_target_id.file.dentry_p = dentry;
    if (rsbac_adf_set_attr(R_CREATE,
                           current->pid,
                           T_DIR,
                           rsbac_target_id,
                           T_FILE,
                           rsbac_new_target_id,
                           A_create_data,
                           rsbac_attribute_value))
      {
        printk(KERN_WARNING
               "rsbac_aef_inode_post_mkdir() [vfs_mkdir() [open_namei() [filp_open() [do_open() [sys_open()]]]]]: rsbac_adf_set_attr() returned error");
      }
  }

static int rsbac_aef_inode_symlink(struct inode *dir,
                                   struct dentry *dentry,
                                   const char *old_name)
  {
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_inode_symlink(): calling ADF\n");
#endif
    rsbac_target_id.dir.device = dir->i_sb->s_dev;
    rsbac_target_id.dir.inode  = dir->i_ino;
    rsbac_target_id.dir.dentry_p = dentry->d_parent;
    rsbac_attribute_value.create_data.target = T_SYMLINK;
    rsbac_attribute_value.create_data.dentry_p = dentry;
    rsbac_attribute_value.create_data.mode = 0;
    rsbac_attribute_value.create_data.device = RSBAC_ZERO_DEV;
    if (!rsbac_adf_request(R_CREATE,
                           current->pid,
                           T_DIR,
                           rsbac_target_id,
                           A_create_data,
                           rsbac_attribute_value))
      return -EPERM;
    else
      return 0;
  }

static void rsbac_aef_inode_post_symlink(struct inode *dir,
                                         struct dentry *dentry,
                                         const char *old_name)
  {
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_target_id_t       rsbac_new_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_inode_post_symlink(): notifying ADF\n");
#endif
    rsbac_target_id.dir.device = dir->i_sb->s_dev;
    rsbac_target_id.dir.inode  = dir->i_ino;
    rsbac_target_id.dir.dentry_p = dentry->d_parent;
    rsbac_attribute_value.create_data.target = T_FILE;
    rsbac_attribute_value.create_data.dentry_p = dentry;
    rsbac_attribute_value.create_data.mode = 0;
    rsbac_attribute_value.create_data.device = RSBAC_ZERO_DEV;
    rsbac_new_target_id.file.device = dentry->d_sb->s_dev;
    rsbac_new_target_id.file.inode  = dentry->d_inode->i_ino;
    rsbac_new_target_id.file.dentry_p = dentry;
    if (rsbac_adf_set_attr(R_CREATE,
                           current->pid,
                           T_DIR,
                           rsbac_target_id,
                           T_SYMLINK,
                           rsbac_new_target_id,
                           A_create_data,
                           rsbac_attribute_value))
      {
        printk(KERN_WARNING
               "rsbac_aef_inode_post_symlink() [vfs_symlink() [open_namei() [filp_open() [do_open() [sys_open()]]]]]: rsbac_adf_set_attr() returned error");
      }
  }

static int rsbac_aef_inode_link(struct dentry *old_dentry,
                                struct inode *dir,
                                struct dentry *new_dentry)
  {
    enum  rsbac_target_t          rsbac_target;
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_inode_link(): calling ADF\n");
#endif
    rsbac_target = T_FILE;
    if (S_ISDIR (old_dentry->d_inode->i_mode))
      rsbac_target = T_DIR;
    else if (S_ISFIFO (old_dentry->d_inode->i_mode))
      rsbac_target = T_FIFO;
    else if (S_ISLNK (old_dentry->d_inode->i_mode))
      rsbac_target = T_SYMLINK;
    rsbac_target_id.dir.device = old_dentry->d_sb->s_dev;
    rsbac_target_id.dir.inode  = old_dentry->d_inode->i_ino;
    rsbac_target_id.dir.dentry_p = old_dentry;
    rsbac_attribute_value.dummy = 0;
    if (!rsbac_adf_request(R_LINK_HARD,
                           current->pid,
                           rsbac_target,
                           rsbac_target_id,
                           A_none,
                           rsbac_attribute_value))
      return -EPERM;
    else
      return 0;
  }

static int rsbac_aef_inode_rename(struct inode *old_dir, struct dentry *old_dentry,
                                  struct inode *new_dir, struct dentry *new_dentry)
  {
    enum  rsbac_target_t          rsbac_target;
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_inode_rename(): calling ADF\n");
#endif
    rsbac_target = T_FILE;
    if (S_ISDIR(old_dentry->d_inode->i_mode))
      rsbac_target = T_DIR;
    else if (S_ISFIFO (old_dentry->d_inode->i_mode))
      rsbac_target = T_FIFO;
    else if (S_ISLNK (old_dentry->d_inode->i_mode))
      rsbac_target = T_SYMLINK;
    rsbac_target_id.file.device = old_dentry->d_sb->s_dev;
    rsbac_target_id.file.inode  = old_dentry->d_inode->i_ino;
    rsbac_target_id.file.dentry_p = old_dentry;
    rsbac_attribute_value.new_dir_dentry_p = new_dentry->d_parent;
    if (!rsbac_adf_request(R_RENAME,
                           current->pid,
                           rsbac_target,
                           rsbac_target_id,
                           A_new_dir_dentry_p,
                           rsbac_attribute_value))
      return -EPERM;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_inode_rename(): calling ADF for WRITE on new dir\n");
#endif
    rsbac_target_id.dir.device = new_dir->i_sb->s_dev;
    rsbac_target_id.dir.inode  = new_dir->i_ino;
    rsbac_target_id.dir.dentry_p = new_dentry->d_parent;
    rsbac_attribute_value.dummy = 0;
    if (!rsbac_adf_request(R_WRITE,
                           current->pid,
                           T_DIR,
                           rsbac_target_id,
                           A_none,
                           rsbac_attribute_value))
      return -EPERM;

    return 0;
  }

static int rsbac_aef_sb_umount(struct vfsmount * mnt, int flags)
  {
    struct super_block * sb = mnt->mnt_sb;
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_sb_umount(): calling ADF for DIR\n");
#endif
    rsbac_target_id.dir.device = sb->s_dev;
    rsbac_target_id.dir.inode  = sb->s_root->d_inode->i_ino;
    rsbac_target_id.dir.dentry_p = sb->s_root;
    rsbac_attribute_value.dummy = 0;
    if (!rsbac_adf_request(R_UMOUNT,
                           current->pid,
                           T_DIR,
                           rsbac_target_id,
                           A_none,
                           rsbac_attribute_value))
      return -EPERM;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_sb_umount(): calling ADF for DEV\n");
#endif
    rsbac_target_id.dev.type = D_block;
    rsbac_target_id.dev.id   = sb->s_dev;
    rsbac_attribute_value.dummy = 0;
    if (!rsbac_adf_request(R_UMOUNT,
                           current->pid,
                           T_DEV,
                           rsbac_target_id,
                           A_none,
                           rsbac_attribute_value))
      return -EPERM;

    if (! (mnt == current->fs->rootmnt && !(flags & MNT_DETACH)) )
      {
#ifdef CONFIG_RSBAC_DEBUG
        if (rsbac_debug_ds)
          printk(KERN_DEBUG "rsbac_aef_sb_umount(): calling rsbac_umount for Device %02u:%02u\n",
                 MAJOR(sb->s_dev), MINOR(sb->s_dev));
#endif
        rsbac_umount(sb, mnt->mnt_mountpoint);
      }
    return 0;
  }

static void rsbac_aef_sb_umount_busy(struct vfsmount * mnt)
  {
    /* RSBAC: reinitialising data structures for this fs (not root fs) */
#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_ds)
      printk(KERN_DEBUG
             "rsbac_aef_sb_umount_busy: umount failed, calling rsbac_mount for Device %02u:%02u\n",
             MAJOR(mnt->mnt_sb->s_dev), MINOR(mnt->mnt_sb->s_dev));
#endif
    rsbac_mount(mnt->mnt_sb, mnt->mnt_mountpoint);
  }

static void rsbac_aef_sb_post_addmount(struct vfsmount * mnt,
                                       struct nameidata * mountpoint_nd)
  {
    /* RSBAC: initialising data structures for this fs (not root fs) */
#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_ds)
      printk(KERN_DEBUG
             "rsbac_aef_sb_post_addmount: calling rsbac_mount for Device %02u:%02u\n",
             MAJOR(mnt->mnt_sb->s_dev), MINOR(mnt->mnt_sb->s_dev));
#endif
    rsbac_mount(mnt->mnt_sb, mnt->mnt_mountpoint);
  }

static int rsbac_aef_sb_mount(char *dev_name, struct nameidata * nd,
                              char *type, unsigned long flags, void *data)
  {
    struct super_block * sb = nd->mnt->mnt_sb;
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_sb_mount(): calling ADF for DIR\n");
#endif
    rsbac_target_id.dir.device = sb->s_dev;
    rsbac_target_id.dir.inode  = sb->s_root->d_inode->i_ino;
    rsbac_target_id.dir.dentry_p = sb->s_root;
    rsbac_attribute_value.dummy = 0;
    if (!rsbac_adf_request(R_MOUNT,
                           current->pid,
                           T_DIR,
                           rsbac_target_id,
                           A_none,
                           rsbac_attribute_value))
      return -EPERM;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_sb_mount(): calling ADF for DEV\n");
#endif
    rsbac_target_id.dev.type = D_block;
    rsbac_target_id.dev.id = sb->s_dev;
    rsbac_attribute_value.dummy = 0;
    if (!rsbac_adf_request(R_MOUNT,
                           current->pid,
                           T_DEV,
                           rsbac_target_id,
                           A_none,
                           rsbac_attribute_value))
      return -EPERM;
    return 0;
  }

static int rsbac_aef_sb_pivotroot(struct nameidata * old_nd,
                                   struct nameidata * new_nd)
  {
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_sb_pivotroot(): calling ADF for MOUNT on put_old\n");
#endif
    rsbac_target_id.dir.device = old_nd->dentry->d_sb->s_dev;
    rsbac_target_id.dir.inode  = old_nd->dentry->d_inode->i_ino;
    rsbac_target_id.dir.dentry_p = old_nd->dentry;
    rsbac_attribute_value.dummy = 0;
    if (!rsbac_adf_request(R_MOUNT,
                           current->pid,
                           T_DIR,
                           rsbac_target_id,
                           A_none,
                           rsbac_attribute_value))
      return -EPERM;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_sb_pivotroot(): calling ADF for MOUNT on root DIR\n");
#endif
    rsbac_target_id.dir.device = current->fs->rootmnt->mnt_sb->s_dev;
    rsbac_target_id.dir.inode  = current->fs->root->d_inode->i_ino;
    rsbac_target_id.dir.dentry_p = current->fs->root;
    rsbac_attribute_value.dummy = 0;
    if (!rsbac_adf_request(R_MOUNT,
                           current->pid,
                           T_DIR,
                           rsbac_target_id,
                           A_none,
                           rsbac_attribute_value))
      return -EPERM;

    return 0;
  }

static int rsbac_aef_inode_getattr(struct vfsmount *mnt, struct dentry *dentry)
  {
    enum  rsbac_target_t          rsbac_target;
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_inode_getattr(): calling ADF\n");
#endif
    rsbac_target = T_FILE;
    if (S_ISDIR (dentry->d_inode->i_mode))
      rsbac_target = T_DIR;
    else if (S_ISFIFO (dentry->d_inode->i_mode))
      rsbac_target = T_FIFO;
    else if (S_ISLNK (dentry->d_inode->i_mode))
      rsbac_target = T_SYMLINK;
    rsbac_target_id.dir.device = dentry->d_sb->s_dev;
    rsbac_target_id.dir.inode  = dentry->d_inode->i_ino;
    rsbac_target_id.dir.dentry_p = dentry;
    rsbac_attribute_value.dummy = 0;
    if (!rsbac_adf_request(R_GET_STATUS_DATA,
                           current->pid,
                           rsbac_target,
                           rsbac_target_id,
                           A_none,
                           rsbac_attribute_value))
      return -EPERM;
    else
      return 0;
  }

static int rsbac_aef_inode_readlink(struct dentry *dentry)
  {
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_inode_readlink(): calling ADF\n");
#endif
    rsbac_target_id.dir.device = dentry->d_sb->s_dev;
    rsbac_target_id.dir.inode  = dentry->d_inode->i_ino;
    rsbac_target_id.dir.dentry_p = dentry;
    rsbac_attribute_value.dummy = 0;
    if (!rsbac_adf_request(R_GET_STATUS_DATA,
                           current->pid,
                           T_SYMLINK,
                           rsbac_target_id,
                           A_none,
                           rsbac_attribute_value))
      return -EPERM;
    else
      return 0;
  }

static int rsbac_aef_file_permission(struct file * file, int mask)
  {
    enum  rsbac_adf_request_t     rsbac_request;
    enum  rsbac_target_t          rsbac_target = T_NONE;
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_file_permission(): calling ADF\n");
#endif
    if(mask & MAY_WRITE)
      rsbac_request = R_WRITE;
    else
      rsbac_request = R_READ;
    #ifdef CONFIG_RSBAC_NET_OBJ_RW
    if(S_ISSOCK(file->f_dentry->d_inode->i_mode))
      {
        #if !defined(CONFIG_RSBAC_NET_OBJ_UNIX)
        if(   file->f_dentry->d_inode->u.socket_i.ops
           && (file->f_dentry->d_inode->u.socket_i.ops->family != AF_UNIX)
          )
        #endif
          {
            rsbac_target = T_NETOBJ;
            rsbac_target_id.netobj.sock_p
              = SOCKET_I(file->f_dentry->d_inode);
            rsbac_target_id.netobj.local_addr = NULL;
            rsbac_target_id.netobj.local_len = 0;
            rsbac_target_id.netobj.remote_addr = NULL;
            rsbac_target_id.netobj.remote_len = 0;
          }
      }
    else
    #endif
    #ifdef CONFIG_RSBAC_RW
    if(   S_ISFIFO(file->f_dentry->d_inode->i_mode)
       /* named FIFO only */
       && (file->f_dentry->d_sb->s_magic != PIPEFS_MAGIC)
      )
      {
        rsbac_target = T_FIFO;
        rsbac_target_id.file.device = file->f_dentry->d_sb->s_dev;
        rsbac_target_id.fifo.inode  = file->f_dentry->d_inode->i_ino;
        rsbac_target_id.fifo.dentry_p = file->f_dentry;
      }
    else
    if(S_ISREG(file->f_dentry->d_inode->i_mode))
      {
        rsbac_target = T_FILE;
        rsbac_target_id.file.device = file->f_dentry->d_sb->s_dev;
        rsbac_target_id.file.inode  = file->f_dentry->d_inode->i_ino;
        rsbac_target_id.file.dentry_p = file->f_dentry;
      }
    else
    if(S_ISBLK(file->f_dentry->d_inode->i_mode))
      {
        rsbac_target = T_DEV;
        rsbac_target_id.dev.type = D_block;
        rsbac_target_id.dev.id = file->f_dentry->d_inode->i_rdev;
      }
    else
    if(S_ISCHR(file->f_dentry->d_inode->i_mode))
      {
        rsbac_target = T_DEV;
        rsbac_target_id.dev.type = D_char;
        rsbac_target_id.dev.id = file->f_dentry->d_inode->i_rdev;
      }
    else
    #endif
    if(S_ISDIR(file->f_dentry->d_inode->i_mode))
      {
        rsbac_target = T_DIR;
        rsbac_target_id.dir.device = file->f_dentry->d_sb->s_dev;
        rsbac_target_id.dir.inode  = file->f_dentry->d_inode->i_ino;
        rsbac_target_id.dir.dentry_p = file->f_dentry;
      }
    rsbac_attribute_value.dummy = 0;
    if (   (rsbac_target != T_NONE)
        && !rsbac_adf_request(rsbac_request,
                              current->pid,
                              rsbac_target,
                              rsbac_target_id,
                              A_none,
                              rsbac_attribute_value))
      return -EPERM;
    else
      return 0;
  }

static int rsbac_aef_task_setgid(gid_t id0, gid_t id1, gid_t id2, int flags)
  {
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_task_setgid(): calling ADF\n");
#endif
    rsbac_target_id.process = current->pid;
    switch(flags)
      {
        case LSM_SETID_RE:
        case LSM_SETID_ID:
        case LSM_SETID_RES:
          if(id0 != (gid_t) -1)
            {
              rsbac_attribute_value.group = id0;
              if (!rsbac_adf_request(R_CHANGE_GROUP,
                                     current->pid,
                                     T_PROCESS,
                                     rsbac_target_id,
                                     A_group,
                                     rsbac_attribute_value))
                return -EPERM;
            }
          break;
        case LSM_SETID_FS:
          break;

        default:
          printk(KERN_WARNING "rsbac_aef_task_setgid(): called with unknown flags value!\n");
          break;
      }
    return 0;
  }

static int rsbac_aef_task_setuid(uid_t id0, uid_t id1, uid_t id2, int flags)
  {
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_task_setuid(): calling ADF\n");
#endif
    rsbac_target_id.process = current->pid;
    switch(flags)
      {
        case LSM_SETID_RE:
          if(id0 != (uid_t) -1)
            {
              rsbac_attribute_value.owner = id0;
              if (!rsbac_adf_request(R_CHANGE_OWNER,
                                     current->pid,
                                     T_PROCESS,
                                     rsbac_target_id,
                                     A_owner,
                                     rsbac_attribute_value))
                return -EPERM;
            }
          #ifdef CONFIG_RSBAC_DAC_OWNER
          if(id1 != (uid_t) -1)
            {
              rsbac_attribute_value.owner = id1;
              if (!rsbac_adf_request(R_CHANGE_DAC_EFF_OWNER,
                                     current->pid,
                                     T_PROCESS,
                                     rsbac_target_id,
                                     A_owner,
                                     rsbac_attribute_value))
                return -EPERM;
            }
          #endif
          break;
        case LSM_SETID_ID:
          rsbac_attribute_value.owner = id0;
          if (!rsbac_adf_request(R_CHANGE_OWNER,
                                 current->pid,
                                 T_PROCESS,
                                 rsbac_target_id,
                                 A_owner,
                                 rsbac_attribute_value))
            return -EPERM;
          #ifdef CONFIG_RSBAC_DAC_OWNER
          if (!rsbac_adf_request(R_CHANGE_DAC_EFF_OWNER,
                                 current->pid,
                                 T_PROCESS,
                                 rsbac_target_id,
                                 A_owner,
                                 rsbac_attribute_value))
            return -EPERM;
          if (!rsbac_adf_request(R_CHANGE_DAC_FS_OWNER,
                                 current->pid,
                                 T_PROCESS,
                                 rsbac_target_id,
                                 A_owner,
                                 rsbac_attribute_value))
            return -EPERM;
          #endif
          break;
        case LSM_SETID_RES:
          if(id0 != (uid_t) -1)
            {
              rsbac_attribute_value.group = id0;
              if (!rsbac_adf_request(R_GET_STATUS_DATA,
                                     current->pid,
                                     T_PROCESS,
                                     rsbac_target_id,
                                     A_group,
                                     rsbac_attribute_value))
                return -EPERM;
            }
          #ifdef CONFIG_RSBAC_DAC_OWNER
          if(id1 != (uid_t) -1)
            {
              rsbac_attribute_value.owner = id1;
              if (!rsbac_adf_request(R_CHANGE_DAC_EFF_OWNER,
                                     current->pid,
                                     T_PROCESS,
                                     rsbac_target_id,
                                     A_owner,
                                     rsbac_attribute_value))
                return -EPERM;
            }
          else
            rsbac_attribute_value.owner = current->euid;
          if (!rsbac_adf_request(R_CHANGE_DAC_FS_OWNER,
                                 current->pid,
                                 T_PROCESS,
                                 rsbac_target_id,
                                 A_owner,
                                 rsbac_attribute_value))
            return -EPERM;
          #endif
          break;
        case LSM_SETID_FS:
          #ifdef CONFIG_RSBAC_DAC_OWNER
          rsbac_attribute_value.owner = id0;
          if (!rsbac_adf_request(R_CHANGE_DAC_FS_OWNER,
                                 current->pid,
                                 T_PROCESS,
                                 rsbac_target_id,
                                 A_owner,
                                 rsbac_attribute_value))
            return -EPERM;
          #endif
          break;

        default:
          printk(KERN_WARNING "rsbac_aef_task_setuid(): called with unknown flags value!\n");
          break;
      }
    return 0;
  }

static int rsbac_aef_task_setgroups(int gidsetsize, gid_t * grouplist)
  {
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_task_setgroups(): calling ADF\n");
#endif
    rsbac_target_id.process = current->pid;
    rsbac_attribute_value.dummy = 0;
    if (!rsbac_adf_request(R_CHANGE_GROUP,
                           current->pid,
                           T_PROCESS,
                           rsbac_target_id,
                           A_none,
                           rsbac_attribute_value))
      return -EPERM;
    else
      return 0;
  }

static int rsbac_aef_task_setrlimit(unsigned int resource, struct rlimit * new_rlim)
  {
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_task_setrlimit(): calling ADF\n");
#endif
    rsbac_target_id.scd = ST_rlimit;
    rsbac_attribute_value.dummy = 0;
    if (!rsbac_adf_request(R_MODIFY_SYSTEM_DATA,
                           current->pid,
                           T_SCD,
                           rsbac_target_id,
                           A_none,
                           rsbac_attribute_value))
      return -EPERM;
    else
      return 0;
  }

static int rsbac_aef_task_kill(struct task_struct * p,
                               struct siginfo * info, int sig)
  {
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#if 0
    if(   (!info || ((unsigned long)info != 1 &&
			(unsigned long)info != 2 && SI_FROMUSER(info)))
       && ((sig != SIGCONT) || (current->session != p->session))
      )
#endif
      {
        if(in_interrupt())
          printk(KERN_DEBUG "signal in interrupt, caller_pid %u, target pid %u, signal %u\n",
                 current->pid, p->pid, sig);
#ifdef CONFIG_RSBAC_DEBUG
        if (rsbac_debug_aef)
          printk(KERN_DEBUG "rsbac_aef_task_kill(): calling ADF\n");
#endif
        rsbac_target_id.process = p->pid;
        rsbac_attribute_value.signal = sig;
        if (!rsbac_adf_request(R_SEND_SIGNAL,
                               current->pid,
                               T_PROCESS,
                               rsbac_target_id,
                               A_signal,
                               rsbac_attribute_value))
          return -EPERM;
      }
    return 0;
  }

#ifdef CONFIG_RSBAC_SYSLOG
static int rsbac_aef_syslog(int type)
  {
    enum  rsbac_adf_request_t     rsbac_request;
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_syslog(): calling ADF\n");
#endif
    switch(type)
      {
        case 2:
        case 3:
          rsbac_request = R_GET_STATUS_DATA;
          break;
        case 4:
        case 5:
        case 6:
        case 7:
        case 8:
          rsbac_request = R_MODIFY_SYSTEM_DATA;
          break;

        default:
          return 0;
      }
    rsbac_target_id.scd = ST_syslog;
    rsbac_attribute_value.dummy = 0;
    if (!rsbac_adf_request(rsbac_request,
                           current->pid,
                           T_SCD,
                           rsbac_target_id,
                           A_signal,
                           rsbac_attribute_value))
      return -EPERM;
    else
      return 0;
  }
#endif

static int rsbac_aef_sysctl(ctl_table * table, int op)
  {
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_sysctl(): calling ADF\n");
#endif
    if(!(op & S_IWUGO))
      return 0;
    switch(table->ctl_name)
      {
        case KERN_NODENAME:
          rsbac_target_id.scd = ST_host_id;
          break;
        case KERN_DOMAINNAME:
          rsbac_target_id.scd = ST_net_id;
          break;
        default:
          rsbac_target_id.scd = ST_other;
      }
    rsbac_attribute_value.dummy = 0;
    if (!rsbac_adf_request(R_MODIFY_SYSTEM_DATA,
                           current->pid,
                           T_SCD,
                           rsbac_target_id,
                           A_signal,
                           rsbac_attribute_value))
      return -EPERM;
    else
      return 0;
  }

static int rsbac_aef_msg_queue_alloc_security(struct msg_queue * msq)
  {
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_msg_queue_alloc_security(): calling ADF\n");
#endif
    rsbac_target_id.ipc.type = I_msg;
    rsbac_target_id.ipc.id.id_nr = 0;
    rsbac_attribute_value.dummy = 0;
    if (!rsbac_adf_request(R_CREATE,
                           current->pid,
                           T_IPC,
                           rsbac_target_id,
                           A_none,
                           rsbac_attribute_value))
      return -EPERM;
    else
      return 0;
  }

static void rsbac_aef_shm_free_security(struct shmid_kernel * shp)
  {
    union rsbac_target_id_t       rsbac_target_id;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_shm_free_security: calling ACI remove_target()\n");
#endif
    rsbac_target_id.ipc.type   = I_shm;
    rsbac_target_id.ipc.id.id_nr  = shp->id;
    rsbac_remove_target(T_IPC, rsbac_target_id);
  }

static int rsbac_aef_shm_alloc_security(struct shmid_kernel * shp)
  {
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_shm_alloc_security(): calling ADF\n");
#endif
    rsbac_target_id.ipc.type = I_shm;
    rsbac_target_id.ipc.id.id_nr = 0;
    rsbac_attribute_value.dummy = 0;
    if (!rsbac_adf_request(R_CREATE,
                           current->pid,
                           T_IPC,
                           rsbac_target_id,
                           A_none,
                           rsbac_attribute_value))
      return -EPERM;
    else
      return 0;
  }

static int rsbac_aef_shm_shmat(struct shmid_kernel * shp, 
                               char *shmaddr, int shmflg)
  {
    enum  rsbac_adf_request_t     rsbac_request;
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_shm_shmat(): calling ADF\n");
#endif
    if ((shmflg & SHM_RDONLY))
      rsbac_request = R_READ_OPEN;
    else
      rsbac_request = R_READ_WRITE_OPEN;
    rsbac_target_id.ipc.type = I_shm;
    rsbac_target_id.ipc.id.id_nr = shp->id;
    rsbac_attribute_value.dummy = 0;
    if (!rsbac_adf_request(rsbac_request,
                           current->pid,
                           T_IPC,
                           rsbac_target_id,
                           A_none,
                           rsbac_attribute_value))
      return -EPERM;
    else
      return 0;
  }

#ifdef CONFIG_RSBAC_IPC_SEM
static int rsbac_aef_sem_alloc_security(struct sem_array * sma)
  {
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

#ifdef CONFIG_RSBAC_DEBUG
    if (rsbac_debug_aef)
      printk(KERN_DEBUG "rsbac_aef_sem_alloc_security(): calling ADF\n");
#endif
    rsbac_target_id.ipc.type = I_sem;
    rsbac_target_id.ipc.id.id_nr = 0;
    rsbac_attribute_value.dummy = 0;
    if (!rsbac_adf_request(R_CREATE,
                           current->pid,
                           T_IPC,
                           rsbac_target_id,
                           A_none,
                           rsbac_attribute_value))
      return -EPERM;
    else
      return 0;
  }
#endif

#ifdef CONFIG_RSBAC_NET_OBJ
static int rsbac_aef_socket_accept(struct socket * sock, struct socket * newsock)
  {
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

    #if !defined(CONFIG_RSBAC_NET_OBJ_UNIX)
    if(   sock->ops
       && (sock->ops->family != AF_UNIX)
      )
    #endif
      {
#ifdef CONFIG_RSBAC_DEBUG
        if (rsbac_debug_aef)
          printk(KERN_DEBUG "rsbac_aef_socket_accept(): calling ADF\n");
#endif
        rsbac_target_id.netobj.sock_p = newsock;
        rsbac_target_id.netobj.local_addr = NULL;
        rsbac_target_id.netobj.local_len = 0;
        rsbac_target_id.netobj.remote_addr = NULL;
        rsbac_target_id.netobj.remote_len = 0;
        rsbac_attribute_value.dummy = 0;
        if (!rsbac_adf_request(R_ACCEPT,
                               current->pid,
                               T_NETOBJ,
                               rsbac_target_id,
                               A_none,
                               rsbac_attribute_value))
          {
#ifdef CONFIG_RSBAC_DEBUG
            if (rsbac_debug_aef)
              printk(KERN_DEBUG "rsbac_aef_socket_accept(): ADF returned NOT_GRANTED\n");
#endif
            return -EPERM;
           }
      }
    return 0;
  }

static void rsbac_aef_socket_post_accept(struct socket * sock, struct socket * newsock)
  {
    union rsbac_target_id_t       rsbac_target_id;
    union rsbac_target_id_t       rsbac_new_target_id;
    union rsbac_attribute_value_t rsbac_attribute_value;

    #if !defined(CONFIG_RSBAC_NET_OBJ_UNIX)
    if(   sock->ops
       && (sock->ops->family != AF_UNIX)
      )
    #endif
      {
        rsbac_target_id.netobj.sock_p = newsock;
        rsbac_target_id.netobj.local_addr = NULL;
        rsbac_target_id.netobj.local_len = 0;
        rsbac_target_id.netobj.remote_addr = NULL;
        rsbac_target_id.netobj.remote_len = 0;
        rsbac_attribute_value.dummy = 0;
        rsbac_new_target_id.dummy = 0;
        if (rsbac_adf_set_attr(R_ACCEPT,
                               current->pid,
                               T_NETOBJ,
                               rsbac_target_id,
                               T_NONE,
                               rsbac_new_target_id,
                               A_none,
                               rsbac_attribute_value))
          {
            printk(KERN_WARNING
                   "rsbac_aef_socket_post_accept(): rsbac_adf_set_attr() returned error\n");
          }
      }
  }
#endif

#ifdef CONFIG_RSBAC_INIT_DELAY
static struct security_operations rsbac_sec_ops = {
#else
static struct security_operations rsbac_sec_ops = {
#endif
  .ptrace = rsbac_aef_ptrace,
  .sb_statfs = rsbac_aef_statfs,
  .inode_follow_link = rsbac_aef_inode_follow_link,
  .inode_create = rsbac_aef_inode_create,
  .inode_post_create = rsbac_aef_inode_post_create,
  .inode_mknod = rsbac_aef_inode_mknod,
  .inode_post_mknod = rsbac_aef_inode_post_mknod,
  .inode_mkdir = rsbac_aef_inode_mkdir,
  .inode_post_mkdir = rsbac_aef_inode_post_mkdir,
  .inode_symlink = rsbac_aef_inode_symlink,
  .inode_post_symlink = rsbac_aef_inode_post_symlink,
  .inode_link = rsbac_aef_inode_link,
  .inode_rename = rsbac_aef_inode_rename,
  .sb_umount = rsbac_aef_sb_umount,
  .sb_umount_busy = rsbac_aef_sb_umount_busy,
  .sb_post_addmount = rsbac_aef_sb_post_addmount,
  .sb_mount = rsbac_aef_sb_mount,
  .sb_pivotroot = rsbac_aef_sb_pivotroot,
  .inode_getattr = rsbac_aef_inode_getattr,
  .inode_readlink = rsbac_aef_inode_readlink,
  .file_permission = rsbac_aef_file_permission,
  .task_setgid = rsbac_aef_task_setgid,
  .task_setuid = rsbac_aef_task_setuid,
  .task_setgroups = rsbac_aef_task_setgroups,
  .task_setrlimit = rsbac_aef_task_setrlimit,
  .task_kill = rsbac_aef_task_kill,
#ifdef CONFIG_RSBAC_SYSLOG
  .syslog = rsbac_aef_syslog,
#endif
  .sysctl = rsbac_aef_sysctl,
  .msg_queue_alloc_security = rsbac_aef_msg_queue_alloc_security,
  .shm_free_security = rsbac_aef_shm_free_security,
  .shm_alloc_security = rsbac_aef_shm_alloc_security,
  #ifdef CONFIG_RSBAC_IPC_SEM
  .sem_alloc_security = rsbac_aef_sem_alloc_security,
  #endif
  #ifdef CONFIG_RSBAC_NET_OBJ
  .socket_accept = rsbac_aef_socket_accept,
  .socket_post_accept = rsbac_aef_socket_post_accept,
  #endif
  .shm_shmat = rsbac_aef_shm_shmat
};

int rsbac_lsm_register(void)
  {
    if (register_security (&rsbac_sec_ops))
      {
        printk (KERN_INFO
                "Failure registering RSBAC with LSM, trying to stack on primary module\n");
        /* try registering with primary module */
        if (mod_reg_security ("RSBAC", &rsbac_sec_ops))
          {
            printk (KERN_ERR "Failure registering RSBAC "
                    "with primary LSM security module, interceptions missing!\n");
            return -EINVAL;
          }
      }
    return 0;
  }
